import tensorflow as tf
import tensorflow_datasets as tfds


def load_train_test_data(strategy, batch_size, buffer_size, tensorflow_seed):
    # Fetch dataset with corresponding information and separate dataset
    datasets, info = tfds.load(name='mnist', with_info=True, as_supervised=True)
    mnist_train, mnist_test = datasets['train'], datasets['test']

    # Distribute data to devices and scale images
    BATCH_SIZE_PER_REPLICA = batch_size
    BATCH_SIZE = BATCH_SIZE_PER_REPLICA * strategy.num_replicas_in_sync

    def scale(image, label):
        image = tf.cast(image, tf.float32)
        image /= 255

        return image, label

    # Ensure that seeds are set and reshuffle_each_iteration is False!
    # https://github.com/tensorflow/tensorflow/issues/38197
    train_dataset = mnist_train.map(scale).cache().shuffle(buffer_size, seed=tensorflow_seed, reshuffle_each_iteration=False).batch(BATCH_SIZE)
    eval_dataset = mnist_test.map(scale).batch(BATCH_SIZE)

    return train_dataset, eval_dataset



import os
from typing import List

import requests
from mlf_core.lint.template_linter import GetLintingFunctionsMeta, TemplateLinter, files_exist_linting
from pkg_resources import parse_version


class PackagePredictionLint(TemplateLinter, metaclass=GetLintingFunctionsMeta):
    def __init__(self, path):
        super().__init__(path)

    def lint(self):
        super().lint_project(self, self.methods)

    def check_dependencies_not_outdated(self) -> bool:
        """
        Check that every dependency from project's requirements.txt is the latest version available at PyPi.

        :return Bool flag that shows code execution went right (used for testing purposes)
        """

        def check_dependencies(filename: str) -> None:
            """
            Check for a given file whether no dependencies are outdated.
            :param filename: Name of the dependency file to parse (either requirements.txt or requirements_dev.txt)
            """
            with open(f"{self.path}/{filename}") as req_file:
                dependencies = [line[:-1].split("==") for line in req_file]
            for dependency in dependencies:
                if len(dependency) == 2:
                    _check_pip_package(dependency[0], dependency[1])

        def _check_pip_package(pip_dependency_name, pip_dependency_version) -> None:
            """
            Query PyPi package information.
            Sends a HTTP GET request to the PyPi remote API.

            :param pip_dependency_name: The name of the dependency
            :param pip_dependency_version: Dependency version used by the user's project
            """
            pip_api_url = f"https://pypi.python.org/pypi/{pip_dependency_name}/json"
            try:
                response = requests.get(pip_api_url, timeout=10)
            except requests.exceptions.Timeout:
                self.warned.append(("cli-python-2", f"PyPi API timed out: {pip_api_url}"))
            except requests.exceptions.ConnectionError:
                self.warned.append(("cli-python-2", f"PyPi API Connection error: {pip_api_url}"))
            else:
                if response.status_code == 200:
                    pip_dep_json = response.json()
                    latest_dependency_version = pip_dep_json["info"]["version"]
                    if parse_version(pip_dependency_version) < parse_version(latest_dependency_version):
                        self.warned.append(
                            (
                                "package-prediction-2",
                                f"Version {pip_dependency_version} of {pip_dependency_name} is not the latest available: "
                                f"{latest_dependency_version}",
                            )
                        )  # noqa: E128
                else:
                    self.failed.append(
                        (
                            "package-prediction-2",
                            f"Could not find pip dependency using the PyPi API: {pip_dependency_name}=={pip_dependency_version}",
                        )
                    )

        # check general dependencies
        check_dependencies("requirements.txt")
        # check development dependencies
        check_dependencies("requirements_dev.txt")
        return True

    def python_files_exist(self) -> None:
        """
        Checks a given project directory for required files.
        Iterates through the templates's directory content and checkmarks files for presence.
        Files that **must** be present::
            'setup.py',
            'setup.cfg',
            'MANIFEST.in',
        Files that *should* be present::
            '.github/workflows/build_package.yml',
            '.github/workflows/publish_package.yml',
            '.github/workflows/flake8.yml',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            '__pycache__'
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [
            ["setup.py"],
            ["setup.cfg"],
            ["MANIFEST.in"],
        ]
        files_warn = [
            [os.path.join(".github", "workflows", "build_package.yml")],
            [os.path.join(".github", "workflows", "publish_package.yml")],
            [os.path.join(".github", "workflows", "run_flake8_linting.yml")],
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists: List = ["__pycache__"]
        files_warn_ifexists: List = []

        files_exist_linting(
            self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, handle="package-prediction"
        )



import os
from pathlib import Path
from typing import List

from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.lint.template_linter import GetLintingFunctionsMeta, TemplateLinter, files_exist_linting
from pkg_resources import parse_version

CWD = os.getcwd()


class MlflowPytorchLint(TemplateLinter, metaclass=GetLintingFunctionsMeta):
    def __init__(self, path):
        super().__init__(path)

    def lint(self):
        super().lint_project(self, self.methods)

    def pytorch_files_exist(self) -> None:
        """
        Checks a given project directory for required files.
        Iterates through the templates's directory content and checkmarks files for presence.
        Files that **must** be present::
            'MLproject',
            'environment.yml',
            'project_slug/'
        Files that *should* be present::
            '.github/workflows/train_cpu.yml',
            '.github/workflows/lint.yml',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            '__pycache__'
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [["MLproject"], ["environment.yml"], [f"{self.project_slug_no_hyphen}/mlf_core/mlf_core.py"]]

        files_warn = [
            [os.path.join(".github", "workflows", "train_cpu.yml")],
            [os.path.join(".github", "workflows", "lint.yml")],
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists: List = ["__pycache__"]
        files_warn_ifexists: List = []

        files_exist_linting(
            self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, handle="mlflow-pytorch"
        )

    def pytorch_reproducibility_seeds(self) -> None:
        """
        Verifies that all CPU and GPU reproducibility settings for Pytorch are enabled
        Required are:
             trainer.deterministic = True,
             trainer.benchmark = False,
             set_general_random_seeds(general_seed),
             set_pytorch_random_seeds(pytorch_seed, num_of_gpus)
        """
        passed_pytorch_reproducibility_seeds = True
        passed_pytorch_reproducibility_seeds_mlf_core = True
        entry_point_file_path = f"{self.path}/{self.project_slug_no_hyphen}/{self.project_slug_no_hyphen}.py"
        mlf_core_file_path = f"{self.path}/{self.project_slug_no_hyphen}/mlf_core/mlf_core.py"
        with open(entry_point_file_path) as f:
            project_slug_entry_point_content = set(map(lambda line: line.strip(), f.readlines()))
        with open(mlf_core_file_path) as f:
            mlf_core_content = set(map(lambda line: line.strip(), f.readlines()))

        expected_lines_pytorch_reproducibility = ["trainer.deterministic = True", "trainer.benchmark = False"]

        expected_lines_pytorch_mlf_core = [
            "torch.manual_seed(seed)",
            "torch.set_deterministic(True)",
            "torch.cuda.manual_seed(seed)",
            "torch.cuda.manual_seed_all(seed)  # For multiGPU",
        ]

        # check for existence of all required entry point files lines for reproducibility
        for expected_line in expected_lines_pytorch_reproducibility:
            if expected_line not in project_slug_entry_point_content:
                passed_pytorch_reproducibility_seeds = False
                self.failed.append(("mlflow-pytorch-2", f"{expected_line} not found in {entry_point_file_path}"))

        # check for existence of required mlf_core lines
        for expected_line in expected_lines_pytorch_mlf_core:
            if expected_line not in mlf_core_content:
                passed_pytorch_reproducibility_seeds_mlf_core = False
                self.failed.append(("mlflow-pytorch-3", f"{expected_line} not found in {entry_point_file_path}"))

        # only pass linting when all seeds and mlf_core settings are set
        if passed_pytorch_reproducibility_seeds and passed_pytorch_reproducibility_seeds_mlf_core:
            self.passed.append(("mlflow-pytorch-2", "All required reproducibility settings enabled."))
            self.passed.append(("mlflow-pytorch-3", "All required reproducibility mlf_core settings enabled."))

    def pytorch_no_atomic_operations(self) -> None:
        """
        Verifies that the project does not use any of the potentially non-deterministic atomicAdd functions.

        There are some PyTorch functions that use CUDA functions that can be a source of nondeterminism.
        One class of such CUDA functions are atomic operations, in particular atomicAdd, which can lead to the order of additions being nondetermnistic.
        Because floating-point addition is not perfectly associative for floating-point operands,
        atomicAdd with floating-point operands can introduce different floating-point rounding errors on each evaluation,
        which introduces a source of nondeterministic variance (aka noise) in the result.
        PyTorch functions that use atomicAdd in the forward kernels include torch.Tensor.index_add_(), torch.Tensor.scatter_add_(), torch.bincount().
        A number of operations have backwards kernels that use atomicAdd, including torch.nn.functional.embedding_bag(),
        torch.nn.functional.ctc_loss(), torch.nn.functional.interpolate(), and many forms of pooling, padding, and sampling.
        There is currently no simple way of avoiding nondeterminism in these functions.
        Additionally, the backward path for repeat_interleave() operates nondeterministically on the CUDA backend because repeat_interleave() is implemented
        using index_select(), the backward path for which is implemented using index_add_(), which is known to operate
        nondeterministically (in the forward direction) on the CUDA backend (see above).

        Source: https://pytorch.org/docs/stable/notes/randomness.html
        """
        atomic_add_functions = [
            "index_add",
            "index_select",
            "scatter_add",
            "bincount",
            "embedding_bag",
            "interpolate",
            "repeat_interleave",
            "histc" "AvgPool3d",
            "AdaptiveAvgPool2d",
            "AdaptiveAvgPool3d",
            "MaxPool3d",
            "AdaptiveMaxPool2d",
            "FractionalMaxPool2d",
            "FractionalMaxPool3d",
            "ReflectionPad1d",
            "ReflectionPad2d",
            "ReplicationPad1d",
            "ReplicationPad2d" "ReplicationPad3d",
            "NLLLoss",
            "CTCLoss",
            "EmbeddingBag"
            # interpolate when called on a CUDA tensor that requires grad and one of the following modes is used: - linear - bilinear - bicubic - trilinear
            # TODO COOKIETEMPLE: This should ideally be solved with a nice regex, which would also help when people have several parameters
            "interpolate('linear')",
            'interpolate("linear")',
            "interpolate(mode='linear')",
            'interpolate(mode="linear")',
            "interpolate('bilinear')",
            'interpolate("bilinear")',
            "interpolate(mode='bilinear')",
            'interpolate(mode="bilinear")',
            "interpolate(bicubic'bicubic')",
            'interpolate("bicubic")',
            "interpolate(mode='bicubic')",
            'interpolate(mode="bicubic")',
            "interpolate(bicubic'trilinear')",
            'interpolate("trilinear")',
            "interpolate(mode='trilinear')",
            'interpolate(mode="trilinear")',
        ]

        verify_method_not_present(self, atomic_add_functions, "mlflow-pytorch-3")


class MlflowTensorflowLint(TemplateLinter, metaclass=GetLintingFunctionsMeta):
    def __init__(self, path):
        super().__init__(path)

    def lint(self):
        super().lint_project(self, self.methods)

    def tensorflow_files_exist(self) -> None:
        """
        Checks a given project directory for required files.
        Iterates through the templates's directory content and checkmarks files for presence.
        Files that **must** be present::
            'MLproject,
            'environment.yml',
            'project_slug_no_hyphen/mlf_core/mlf_core.py',
        Files that *should* be present::
            '.github/workflows/train_cpu.yml',
            '.github/workflows/lint.yml',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            '__pycache__'
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [["MLproject"], ["environment.yml"], [f"{self.project_slug_no_hyphen}/mlf_core/mlf_core.py"]]
        files_warn = [
            [os.path.join(".github", "workflows", "train_cpu.yml")],
            [os.path.join(".github", "workflows", "lint.yml")],
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists: List = ["__pycache__"]
        files_warn_ifexists: List = []

        files_exist_linting(
            self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, handle="mlflow-tensorflow"
        )

    def tensorflow_reproducibility_seeds(self) -> None:
        """
        Verifies that all CPU and GPU reproducibility settings for Tensorflow are enabled
        Required are:
        set_general_random_seeds(dict_args["general_seed"])
        set_tensorflow_random_seeds(dict_args["tensorflow_seed"])
        def set_tensorflow_random_seeds(seed):
            tf.random.set_seed(seed)
            tf.config.threading.set_intra_op_parallelism_threads = 1  # CPU only -> https://github.com/NVIDIA/tensorflow-determinism
            tf.config.threading.set_inter_op_parallelism_threads = 1  # CPU only
            os.environ['TF_DETERMINISTIC_OPS'] = '1'
        """
        passed_tensorflow_reproducibility_seeds = True
        entry_point_file_path = f"{self.path}/{self.project_slug_no_hyphen}/{self.project_slug_no_hyphen}.py"
        with open(entry_point_file_path) as f:
            project_slug_entry_point_content = list(map(lambda line: line.strip(), f.readlines()))

        expected_lines_tensorflow_reproducibility = [
            'MLFCore.set_general_random_seeds(dict_args["general_seed"])',
            'MLFCore.set_tensorflow_random_seeds(dict_args["tensorflow_seed"])',
        ]

        for expected_line in expected_lines_tensorflow_reproducibility:
            if expected_line not in project_slug_entry_point_content:
                passed_tensorflow_reproducibility_seeds = False
                self.failed.append(("mlflow-tensorflow-2", f"{expected_line} not found in {entry_point_file_path}"))

        if passed_tensorflow_reproducibility_seeds:
            self.passed.append(("mlflow-tensorflow-2", "All required reproducibility settings enabled."))

        # TODO COOKIETEMPLE: Investigate whether there is a reasonable way of linting for
        # train_dataset = mnist_train.map(scale).cache().shuffle(buffer_size, seed=tensorflow_seed, reshuffle_each_iteration=False).batch(BATCH_SIZE)

    def tensorflow_non_deterministic_functions(self) -> None:
        """
        Verifies that no non-deterministic functions of Tensorflow are used.
        """
        non_deterministic_tf_functions = [
            "softmax_cross_entropy_with_logits",
            "sparse_softmax_cross_entropy_with_logits",
        ]

        verify_method_not_present(self, non_deterministic_tf_functions, "mlflow-tensorflow-3")


class MlflowXGBoostLint(TemplateLinter, metaclass=GetLintingFunctionsMeta):
    def __init__(self, path):
        super().__init__(path)

    def lint(self):
        super().lint_project(self, self.methods)

    def xgboost_files_exist(self) -> None:
        """
        Checks a given project directory for required files.
        Iterates through the templates's directory content and checkmarks files for presence.
        Files that **must** be present::
            'MLproject',
            'environment.yml',
            'project_slug/mlf_core/mlf_core.py'
        Files that *should* be present::
            '.github/workflows/train_cpu.yml',
            '.github/workflows/lint.yml',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            '__pycache__'
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [["MLproject"], ["environment.yml"], [f"{self.project_slug_no_hyphen}/mlf_core/mlf_core.py"]]
        files_warn = [
            [os.path.join(".github", "workflows", "train_cpu.yml")],
            [os.path.join(".github", "workflows", "lint.yml")],
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists: List = ["__pycache__"]
        files_warn_ifexists: List = []

        files_exist_linting(
            self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, handle="mlflow-xgboost"
        )

    def xgboost_reproducibility_seeds(self) -> None:
        """
        Verifies that all CPU and GPU reproducibility settings for XGBoost are enabled
        """
        passed_xgboost_reproducibility_seeds = True
        entry_point_file_path = f"{self.path}/{self.project_slug_no_hyphen}/{self.project_slug_no_hyphen}.py"
        with open(entry_point_file_path) as f:
            project_slug_entry_point_content = list(map(lambda line: line.strip(), f.readlines()))

        expected_lines_xgboost_reproducibility = [
            'MLFCore.set_general_random_seeds(dict_args["general_seed"])',
            'MLFCore.set_xgboost_random_seeds(dict_args["xgboost_seed"], param)',
        ]

        for expected_line in expected_lines_xgboost_reproducibility:
            if expected_line not in project_slug_entry_point_content:
                passed_xgboost_reproducibility_seeds = False
                self.failed.append(("mlflow-xgboost-2", f"{expected_line} not found in {entry_point_file_path}"))

        if passed_xgboost_reproducibility_seeds:
            self.passed.append(("mlflow-xgboost-2", "All required reproducibility settings enabled."))

    def xgboost_version(self) -> None:
        """
        Verifies that the XGBoost version is at least 1.1.0, since reproducibility cannot be guaranteed elsewise.
        """
        # Verify that XGBoost version is greater than 1.1.0
        conda_env = load_yaml_file(f"{self.path}/environment.yml")
        conda_only = list(filter(lambda dep: "::" in dep, conda_env["dependencies"]))
        pip_only = list(filter(lambda dep: isinstance(dep, dict), conda_env["dependencies"]))[0]["pip"]

        for dependency in conda_only:
            if "xgboost" in dependency:
                split = dependency.split("==")
                current_version = parse_version(split[-1])
                if current_version < parse_version("1.1.0"):
                    self.failed.append(
                        (
                            "mlflow-xgboost-3",
                            f"XGBoost version {current_version} is not at least 1.1.0. Reproducibility cannot be guaranteed.",
                        )
                    )

        for dependency in pip_only:
            if "xgboost" in dependency:
                split = dependency.split("==")
                current_version = parse_version(split[-1])
                if current_version < parse_version("1.1.0"):
                    self.failed.append(
                        (
                            "mlflow-xgboost-3",
                            f"XGBoost version {current_version} is not at least 1.1.0. Reproducibility cannot be guaranteed.",
                        )
                    )

    def xgboost_no_all_reduce(self) -> None:
        """
        Verifies that all_reduce is not used.
        https://github.com/dmlc/xgboost/issues/5023
        """
        all_reduce_functions = ["all_reduce"]
        verify_method_not_present(self, all_reduce_functions, "mlflow-xgboost-4")


def verify_method_not_present(calling_class: TemplateLinter, functions_to_check: list, linting_code: str):
    """
    Checks all Python files for a list of strings/functions, which should not be present.
    :param calling_class: The TemplateLinter subclass. Required to append the found errors correctly.
    :param functions_to_check: The list of functions, which are not allowed to be present in any Python file
    :param linting_code: A linting code build from the handle and the error code e.g. mlflow-pytorch-1
    """
    # We should only expect those functions in all *.py files, so only read those
    for root, _, files in os.walk(f"{calling_class.path}/{calling_class.project_slug_no_hyphen}"):
        for file in files:
            if file.endswith(".py"):
                file_to_check_full_path = Path(root) / file
                with open(file_to_check_full_path) as f:
                    content_stripped = list(map(lambda line: line.strip(), f.readlines()))
                    for function in functions_to_check:
                        for line_code in content_stripped:
                            if function in line_code:
                                calling_class.failed.append(
                                    (
                                        linting_code,
                                        f"{function} found in {file_to_check_full_path} operates non-deterministically.",
                                    )
                                )



import logging
import os

from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.util.dict_util import is_nested_dictionary
from rich import print
from rich.box import HEAVY_HEAD
from rich.console import Console
from rich.style import Style
from rich.table import Table

log = logging.getLogger(__name__)


class TemplateLister:
    """
    A class responsible for listing all available mlf-core templates in a nice layout
    """

    WD = os.path.dirname(__file__)
    TEMPLATES_PATH = f"{WD}/../create/templates"

    @classmethod
    def list_available_templates(cls) -> None:
        """
        Displays all available templates to stdout in nicely formatted yaml format.
        Omits long descriptions.
        """
        log.debug(f"Reading available_templates.yml at {cls.TEMPLATES_PATH}/available_templates.yml")
        available_templates = load_yaml_file(f"{cls.TEMPLATES_PATH}/available_templates.yml")
        print("[bold blue]Run [green]mlf-core info [blue]for long descriptions of your template of interest")
        print()

        # What we want to have are lists like
        # [['name', 'handle', 'short description', 'available libraries', 'version'], ['name', 'handle', 'short description', 'available libraries', 'version']]
        log.debug("Building list table.")
        templates_to_tabulate = []
        for language in available_templates.values():
            for val in language.values():
                # has a subdomain -> traverse dictionary a level deeper
                if is_nested_dictionary(val):
                    for val_2 in val.values():
                        templates_to_tabulate.append(
                            [
                                val_2["name"],
                                val_2["handle"],
                                val_2["short description"],
                                val_2["available libraries"],
                                val_2["version"],
                            ]
                        )
                else:
                    templates_to_tabulate.append(
                        [
                            val["name"],
                            val["handle"],
                            val["short description"],
                            val["available libraries"],
                            val["version"],
                        ]
                    )

        table = Table(
            title="[bold]All available mlf-core templates",
            title_style="blue",
            header_style=Style(color="blue", bold=True),
            box=HEAVY_HEAD,
        )

        table.add_column("Name", justify="left", style="green", no_wrap=True)
        table.add_column("Handle", justify="left")
        table.add_column("Short Description", justify="left")
        table.add_column("Available Libraries", justify="left")
        table.add_column("Version", justify="left")

        for template in templates_to_tabulate:
            table.add_row(f"[bold]{template[0]}", template[1], template[2], template[3], template[4])

        console = Console()
        log.debug("Printing list table.")
        console.print(table)



import os
from dataclasses import dataclass
from pathlib import Path

from mlf_core.common.version import load_mlf_core_template_version
from mlf_core.create.domains.mlf_core_template_struct import MlfcoreTemplateStruct
from mlf_core.create.github_support import prompt_github_repo
from mlf_core.create.template_creator import TemplateCreator
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core


@dataclass
class TemplateStructPackage(MlfcoreTemplateStruct):
    """
    PACKAGE-PREDICTION
    """

    framework: str = ""  # which machine learning framework to use


class PackageCreator(TemplateCreator):
    def __init__(self):
        self.package_struct = TemplateStructPackage(domain="package")
        super().__init__(self.package_struct)
        self.WD_Path = Path(os.path.dirname(__file__))
        self.TEMPLATES_PACKAGE_PATH = f"{self.WD_Path.parent}/templates/package"

        '"" TEMPLATE VERSIONS ""'
        self.PACKAGE_PREDICTION_TEMPLATE_VERSION = load_mlf_core_template_version(
            "package-prediction", self.AVAILABLE_TEMPLATES_PATH
        )

    def create_template(self, path: Path, dot_mlf_core):
        """
        Handles the package domain. Prompts the user for the language, general and domain specific options.
        """

        self.package_struct.language = mlf_core_questionary_or_dot_mlf_core(
            function="select",
            question="Choose the project's primary purpose",
            choices=["prediction"],
            default="prediction",
            dot_mlf_core=dot_mlf_core,
            to_get_property="language",
        )

        # prompt the user to fetch general template configurations
        super().prompt_general_template_configuration(dot_mlf_core)

        # switch case statement to prompt the user to fetch template specific configurations
        switcher = {
            "prediction": self.package_prediction_options,
        }
        switcher.get(self.package_struct.language)(dot_mlf_core)  # type: ignore

        (
            self.package_struct.is_github_repo,
            self.package_struct.is_repo_private,
            self.package_struct.is_github_orga,
            self.package_struct.github_orga,
        ) = prompt_github_repo(dot_mlf_core)

        if self.package_struct.is_github_orga:
            self.package_struct.github_username = self.package_struct.github_orga
        # create the chosen and configured template
        super().create_template_without_subdomain(self.TEMPLATES_PACKAGE_PATH)

        # switch case statement to fetch the template version
        switcher_version = {
            "package": self.PACKAGE_PREDICTION_TEMPLATE_VERSION,
        }
        self.package_struct.template_version, self.package_struct.template_handle = (
            switcher_version.get(self.package_struct.language),  # type: ignore
            f"package-{self.package_struct.language.lower()}",  # type: ignore
        )

        # perform general operations like creating a GitHub repository and general linting
        super().process_common_operations(
            path=Path(path).resolve(),
            domain="package",
            language=self.package_struct.language,  # type: ignore
            dot_mlf_core=dot_mlf_core,
        )

    def package_prediction_options(self, dot_mlf_core):
        """ Prompts for package-prediction specific options and saves them into the PackageTemplateStruct """
        self.package_struct.framework = mlf_core_questionary_or_dot_mlf_core(
            function="select",
            question="Choose a framework",
            choices=["pytorch", "tensorflow", "xgboost"],
            default="pytorch",
            dot_mlf_core=dot_mlf_core,
            to_get_property="framework",
        )



import logging
import sys
from typing import Union

import questionary
from prompt_toolkit.styles import Style
from rich import print

log = logging.getLogger(__name__)


mlf_core_style = Style(
    [
        ("qmark", "fg:#0000FF bold"),  # token in front of the question
        ("question", "bold"),  # question text
        ("answer", "fg:#008000 bold"),  # submitted answer text behind the question
        ("pointer", "fg:#0000FF bold"),  # pointer used in select and checkbox prompts
        ("highlighted", "fg:#0000FF bold"),  # pointed-at choice in select and checkbox prompts
        ("selected", "fg:#008000"),  # style for a selected item of a checkbox
        ("separator", "fg:#cc5454"),  # separator in lists
        ("instruction", ""),  # user instructions for select, rawselect, checkbox
        ("text", ""),  # plain text
        ("disabled", "fg:#FF0000 italic"),  # disabled choices for select and checkbox prompts
    ]
)


def mlf_core_questionary_or_dot_mlf_core(
    function: str,
    question: str,
    choices: list = None,
    default: str = None,
    dot_mlf_core: dict = None,
    to_get_property: str = None,
) -> Union[str, bool]:
    """
    Custom selection based on Questionary. Handles keyboard interrupts and default values.

    :param function: The function of questionary to call (e.g. select or text). See https://github.com/tmbo/questionary for all available functions.
    :param choices: List of all possible choices.
    :param question: The question to prompt for. Should not include default values or colons.
    :param default: A set default value, which will be chosen if the user does not enter anything.
    :param dot_mlf_core: A dictionary, which contains the whole .mlf_core.yml content
    :param to_get_property: A key, which must be in the dot_mlf_core file, which is used to fetch the read in value from the .mlf_core.yml file
    :return: The chosen answer.
    """
    # First check whether a dot_mlf_core was passed and whether it contains the desired property -> return it if so
    try:
        if dot_mlf_core:
            if to_get_property in dot_mlf_core:
                return dot_mlf_core[to_get_property]
    except KeyError:
        log.debug(
            f".mlf_core.yml file was passed when creating a project, but key {to_get_property}"
            f" does not exist in the dot_mlf_core dictionary! Assigning default {default} to {to_get_property}."
        )
        return default  # type: ignore

    # There is no .mlf_core.yml file aka dot_mlf_core dict passed -> ask for the properties
    answer = ""
    try:
        if function == "select":
            if default not in choices:  # type: ignore
                log.debug(f"Default value {default} is not in the set of choices!")
            answer = getattr(questionary, function)(f"{question}: ", choices=choices, style=mlf_core_style).unsafe_ask()
        elif function == "password":
            while not answer or answer == "":
                answer = getattr(questionary, function)(f"{question}: ", style=mlf_core_style).unsafe_ask()
        elif function == "text":
            if not default:
                log.debug(
                    "Tried to utilize default value in questionary prompt, but is None! Please set a default value."
                )
                default = ""
            answer = getattr(questionary, function)(f"{question} [{default}]: ", style=mlf_core_style).unsafe_ask()
        elif function == "confirm":
            default_value_bool = True if default == "Yes" or default == "yes" else False
            answer = getattr(questionary, function)(
                f"{question} [{default}]: ", style=mlf_core_style, default=default_value_bool
            ).unsafe_ask()
        else:
            log.debug(f"Unsupported questionary function {function} used!")

    except KeyboardInterrupt:
        print("[bold red] Aborted!")
        sys.exit(1)
    if answer is None or answer == "":
        answer = default  # type: ignore

    return answer



from pathlib import Path

from mlf_core.create.domains.mlflow_creator import MlflowCreator
from mlf_core.create.domains.package_creator import PackageCreator
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core


def choose_domain(path: Path, domain: str or None, dot_mlf_core: dict = None):  # type: ignore
    """
    Prompts the user for the template domain.
    Creates the .mlf_core file.
    Prompts the user whether or not to create a Github repository

    :param path: The path, the project should be created at
    :param domain: Template domain
    :param dot_mlf_core: Dictionary created from the .mlf_core.yml file. None if no .mlf_core.yml file was used.
    """
    if not domain:
        domain = mlf_core_questionary_or_dot_mlf_core(
            function="select",
            question="Choose the project's domain",
            choices=["mlflow", "package"],
            default="mlflow",
            dot_mlf_core=dot_mlf_core,
            to_get_property="domain",
        )

    switcher = {"mlflow": MlflowCreator, "package": PackageCreator}

    creator_obj = switcher.get(domain.lower())()  # type: ignore
    creator_obj.create_template(path, dot_mlf_core)  # type: ignore



import os
import sys
from dataclasses import dataclass

import click
import numpy as np
import xgboost as xgb
from rich import print, traceback

WD = os.path.dirname(__file__)


@click.command()
@click.option('-i', '--input', required=True, type=str, help='Path to data file to predict.')
@click.option('-m', '--model', type=str, help='Path to an already trained XGBoost model. If not passed a default model will be loaded.')
@click.option('-c/-nc', '--cuda/--no-cuda', type=bool, default=False, help='Whether to enable cuda or not')
@click.option('-o', '--output', type=str, help='Path to write the output to')
def main(input: str, model: str, cuda: bool, output: str):
    """Command-line interface for {{ cookiecutter.project_name }}"""

    print(r"""[bold blue]
        {{ cookiecutter.project_name }}
        """)

    print('[bold blue]Run [green]{{ cookiecutter.project_name }} --help [blue]for an overview of all commands\n')
    if not model:
        model = get_xgboost_model(f'{WD}/models/xgboost_test_model.xgb')
    else:
        model = get_xgboost_model(model)
    if cuda:
        model.set_param({'predictor': 'gpu_predictor'})
    print('[bold blue] Parsing data')
    data_to_predict = parse_data_to_predict(input)
    print('[bold blue] Performing predictions')
    predictions = np.round(model.predict(data_to_predict.DM))
    print(predictions)
    if output:
        print(f'[bold blue]Writing predictions to {output}')
        write_results(predictions, output)


@dataclass
class Dataset:
    X: np.ndarray
    y: list
    DM: xgb.DMatrix
    gene_names: list
    sample_names: list


def parse_data_to_predict(path_to_data_to_predict: str) -> Dataset:
    """
    Parses the data to predict and returns a full Dataset include the DMatrix
    :param path_to_data_to_predict: Path to the data on which predictions should be performed on
    """
    X = []
    y = []
    gene_names = []
    sample_names = []
    with open(path_to_data_to_predict, "r") as file:
        all_runs_info = next(file).split("\n")[0].split("\t")[2:]
        for run_info in all_runs_info:
            split_info = run_info.split("_")
            y.append(int(split_info[0]))
            sample_names.append(split_info[1])
        for line in file:
            split = line.split("\n")[0].split("\t")
            X.append([float(x) for x in split[2:]])
            gene_names.append(split[:2])

    X = [list(i) for i in zip(*X)]

    X_np = np.array(X)
    DM = xgb.DMatrix(X_np, label=y)

    return Dataset(X_np, y, DM, gene_names, sample_names)


def write_results(predictions: np.ndarray, path_to_write_to) -> None:
    """
    Writes the predictions into a human readable file.
    :param predictions: Predictions as a numpy array
    :param path_to_write_to: Path to write the predictions to
    """
    np.savetxt(path_to_write_to, predictions, delimiter=',')


def get_xgboost_model(path_to_xgboost_model: str):
    """
    Fetches the model of choice and creates a booster from it.
    :param path_to_xgboost_model: Path to the xgboost model1
    """
    model = xgb.Booster()
    model.load_model(os.path.abspath(path_to_xgboost_model))

    return model


if __name__ == "__main__":
    traceback.install()
    sys.exit(main())  # pragma: no cover



import os

import tensorflow as tf


def train(model, epochs, train_dataset):
    # Define the checkpoint directory to store the checkpoints
    checkpoint_dir = './training_checkpoints'
    checkpoint_prefix = os.path.join(checkpoint_dir, "ckpt_{epoch}")

    callbacks = [
        tf.keras.callbacks.TensorBoard(log_dir='./logs'),
        tf.keras.callbacks.ModelCheckpoint(filepath=checkpoint_prefix, save_weights_only=True)
    ]

    model.fit(train_dataset, epochs=epochs, callbacks=callbacks)
    model.load_weights(tf.train.latest_checkpoint(checkpoint_dir))


def test(model, eval_dataset):
    eval_loss, eval_acc = model.evaluate(eval_dataset)

    return eval_loss, eval_acc



import os
import shutil


def remove(filepath):
    if os.path.isfile(filepath):
        os.remove(filepath)
    elif os.path.isdir(filepath):
        shutil.rmtree(filepath)


cli_path = '{{ cookiecutter.project_slug_no_hyphen }}/{{ cookiecutter.project_slug_no_hyphen }}'

pytorch_files = [f'{cli_path}/cli_pytorch.py']

tensorflow_files = [f'{cli_path}/cli_tensorflow.py']

xgboost_files = [f'{cli_path}/cli_xgboost.py',
                 '{{ cookiecutter.project_slug_no_hyphen }}/data/xgboost_test_data.tsv',
                 '{{ cookiecutter.project_slug_no_hyphen }}/models/xgboost_test_model.xgb']

is_pytorch = '{{ cookiecutter.framework }}' == 'pytorch'
is_tensorflow = '{{ cookiecutter.framework }}' == 'tensorflow'
is_xgboost = '{{ cookiecutter.framework }}' == 'xgboost'

if not is_pytorch:
    for file in pytorch_files:
        remove(os.path.join(os.getcwd(), file))

if not is_tensorflow:
    for file in tensorflow_files:
        remove(os.path.join(os.getcwd(), file))

if not is_xgboost:
    for file in xgboost_files:
        remove(os.path.join(os.getcwd(), file))



import os
import time
from argparse import ArgumentParser

import mlflow
import mlflow.tensorflow
import tensorflow as tf
from data_loading.data_loader import load_train_test_data
from model.model import create_model
from rich import print, traceback
from training.train import test, train

from mlf_core.mlf_core import MLFCore


def start_training():
    parser = ArgumentParser(description='Tensorflow example')
    parser.add_argument(
        '--cuda',
        type=bool,
        default=True,
        help='Enable or disable CUDA support',
    )
    parser.add_argument(
        '--max_epochs',
        type=int,
        default=10,
        help='Number of epochs to train',
    )
    parser.add_argument(
        '--general-seed',
        type=int,
        default=0,
        help='General Python, Python random and Numpy seed.',
    )
    parser.add_argument(
        '--tensorflow-seed',
        type=int,
        default=0,
        help='Tensorflow specific random seed.',
    )
    parser.add_argument(
        '--batch-size',
        type=int,
        default=64,
        help='Input batch size for training and testing',
    )
    parser.add_argument(
        '--buffer-size',
        type=int,
        default=10000,
        help='Buffer size for Mirrored Training',
    )
    parser.add_argument(
        '--lr',
        type=float,
        default=0.01,
        help='Learning rate',
    )
    args = parser.parse_args()
    dict_args = vars(args)
    # Disable GPU support if no GPUs are supposed to be used
    if not dict_args['cuda']:
        tf.config.set_visible_devices([], 'GPU')

    with mlflow.start_run():
        # Enable the logging of all parameters, metrics and models to mlflow and Tensorboard
        mlflow.autolog(1)

        # Log hardware and software
        MLFCore.log_sys_intel_conda_env()

        # Fix all random seeds and Tensorflow specific reproducibility settings
        MLFCore.set_general_random_seeds(dict_args["general_seed"])
        MLFCore.set_tensorflow_random_seeds(dict_args["tensorflow_seed"])

        # Use Mirrored Strategy for multi GPU support
        strategy = tf.distribute.MirroredStrategy()
        print(f'[bold blue]Number of devices: {strategy.num_replicas_in_sync}')

        # Fetch and prepare dataset
        train_dataset, eval_dataset = load_train_test_data(strategy, dict_args['batch_size'], dict_args['buffer_size'], dict_args['tensorflow_seed'])

        # TODO MLF-CORE: Enable input data logging
        # MLFCore.log_input_data('data/')

        with strategy.scope():
            # Define model and compile model
            model = create_model(input_shape=(28, 28, 1))
            model.compile(loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
                          optimizer=tf.keras.optimizers.Adam(learning_rate=dict_args['lr']),
                          metrics=['accuracy'])

            # Train and evaluate the trained model
            runtime = time.time()
            train(model, dict_args['max_epochs'], train_dataset)
            eval_loss, eval_acc = test(model, eval_dataset)
            print(f'Test loss: {eval_loss}, Test Accuracy: {eval_acc}')

            device = 'GPU' if dict_args['cuda'] else 'CPU'
            print(f'[bold green]{device} Run Time: {str(time.time() - runtime)} seconds')

            print(f'[bold blue]\nLaunch TensorBoard with:\ntensorboard --logdir={os.path.join(mlflow.get_artifact_uri(), "tensorboard_logs", "train")}')


if __name__ == '__main__':
    traceback.install()
    print(f'Num GPUs Available: {len(tf.config.experimental.list_physical_devices("GPU"))}')

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # Filtering out any Warnings messages

    start_training()



#!/usr/bin/env python
"""
Synchronise a project TEMPLATE branch with the template.
"""
import fnmatch
import logging
import os
import shutil
import sys
import tempfile
from configparser import ConfigParser, NoSectionError
from distutils.dir_util import copy_tree
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Tuple

import git
from github import Github, GithubException
from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.common.version import load_mlf_core_template_version, load_project_template_version_and_handle
from mlf_core.create.github_support import create_sync_secret, decrypt_pat, load_github_username
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from packaging import version
from rich import print

from mlf_core.config.config import ConfigCommand
from mlf_core.create.create import choose_domain

log = logging.getLogger(__name__)


class TemplateSync:
    """
    Hold syncing information and results.

    project_dir (str): The path to the mlf-core project root directory
    from_branch (str): Original branch
    project_dir (str): Path to target project directory
    from_branch (str): Original branch
    original_branch (str): Repo branch that was checked out before we started.
    made_changes (bool): Whether making the new template project introduced any changes
    gh_username (str): GitHub username
    patch_update (bool): Whether a patch update was found for the template or not
    minor_update (bool): Whether a minor update was found for the template or not
    major_update (bool): Whether a major update was found for the template or not
    repo_owner (str): Owner of the repo (either orga name or personal github username)
    """

    def __init__(
        self,
        project_dir,
        new_template_version,
        from_branch=None,
        gh_username=None,
        token=None,
        major_update=False,
        minor_update=False,
        patch_update=False,
    ):
        self.project_dir = os.path.abspath(project_dir)
        self.from_branch = from_branch
        self.original_branch = None
        self.made_changes = False
        self.gh_pr_returned_data = {}
        self.major_update = major_update
        self.minor_update = minor_update
        self.patch_update = patch_update
        self.gh_username = gh_username if gh_username else load_github_username()
        self.token = token if token else decrypt_pat()
        self.dot_mlf_core = {}
        self.repo_owner = self.gh_username
        self.new_template_version = new_template_version
        self.github = Github(self.token)
        self.blacklisted_globs = []

    def sync(self):
        """
        Sync the mlf-core project
        """
        self.inspect_sync_dir()
        # get blacklisted files on current working branch
        self.blacklisted_globs = self.get_blacklisted_sync_globs()
        self.checkout_template_branch()
        self.delete_template_branch_files()
        self.make_template_project()
        self.commit_template_changes()

        # Push and make a pull request
        if self.made_changes:
            try:
                self.push_template_branch()
                self.make_pull_request()
            except Exception as e:
                self.reset_target_dir()
                print(f"[bold red]{e}")
                sys.exit(1)

        self.reset_target_dir()

        if not self.made_changes:
            print("[bold blue]No changes made to TEMPLATE - sync complete")

    def inspect_sync_dir(self):
        """
        Examines target directory to sync, verifies that it is a git repository and ensures that there are no uncommitted changes.
        """
        if not os.path.exists(os.path.join(str(self.project_dir), ".mlf_core.yml")):
            print(f"[bold red]{self.project_dir} does not appear to contain a .mlf_core.yml file. Did you delete it?")
            sys.exit(1)
        # store .mlf_core.yml content for later reuse in the dry create run
        self.dot_mlf_core = load_yaml_file(os.path.join(str(self.project_dir), ".mlf_core.yml"))
        log.debug(f"Loaded .mlf_core.yml file content. Content is: {self.dot_mlf_core}")
        # Check that the project_dir is a git repo
        try:
            self.repo = git.Repo(self.project_dir)
        except git.exc.InvalidGitRepositoryError:
            print(f"[bold red]{self.project_dir} does not appear to be a git repository.")
            sys.exit(1)

        # get current branch so we can switch back later
        self.original_branch = self.repo.active_branch.name
        print(f"[bold blue]Original Project repository branch is {self.original_branch}")

        # Check to see if there are uncommitted changes on current branch
        if self.repo.is_dirty(untracked_files=True):
            print(
                "[bold red]Uncommitted changes found in Project directory!\nPlease commit these before running mlf-core sync"
            )
            sys.exit(1)

    def checkout_template_branch(self):
        """
        Try to check out the origin/TEMPLATE in a new TEMPLATE branch.
        If this fails, try to check out an existing local TEMPLATE branch.
        """
        try:
            self.from_branch = self.repo.active_branch.name
        except git.exc.GitCommandError as e:
            print(f"[bold red]Could not find active repo branch:\n{e}")
        # Try to check out the `TEMPLATE` branch
        try:
            self.repo.git.checkout("origin/TEMPLATE", b="TEMPLATE")
        except git.exc.GitCommandError:
            # Try to check out an existing local branch called TEMPLATE
            try:
                self.repo.git.checkout("TEMPLATE")
            except git.exc.GitCommandError:
                print('[bold red]Could not check out branch "origin/TEMPLATE" or "TEMPLATE"')
                sys.exit(1)

    def delete_template_branch_files(self):
        """
        Delete all files in the TEMPLATE branch
        """
        # Delete everything
        print("[bold blue]Deleting all files in TEMPLATE branch")
        for the_file in os.listdir(self.project_dir):
            if the_file == ".git":
                log.debug("Found .git directory. Skipping deleting it.")
                continue
            file_path = os.path.join(self.project_dir, the_file)
            try:
                if os.path.isfile(file_path):
                    log.debug(f"Deleting file {file_path}")
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    log.debug(f"Deleting directory {file_path}")
                    shutil.rmtree(file_path)
            except Exception as e:
                print(f"[bold red]{e}")
                sys.exit(1)

    def make_template_project(self):
        """
        Delete all files and make a fresh template.
        """
        print("[bold blue]Creating a new template project.")
        # dry create run from mlf-core in tmp directory
        with tempfile.TemporaryDirectory() as tmpdirname:
            old_cwd = str(Path.cwd())
            log.debug(f"Saving current working directory {old_cwd}.")
            os.chdir(tmpdirname)
            log.debug(f"Changed directory to {tmpdirname}.")
            log.debug(f"Calling choose_domain with {self.dot_mlf_core}.")
            choose_domain(path=Path.cwd(), domain=None, dot_mlf_core=self.dot_mlf_core)
            # copy into the cleaned TEMPLATE branch's project directory
            log.debug(f"Copying created template into {self.project_dir}.")
            copy_tree(os.path.join(tmpdirname, self.dot_mlf_core["project_slug"]), str(self.project_dir))
            log.debug(f"Changing directory back to {old_cwd}.")
            os.chdir(old_cwd)

    def commit_template_changes(self):
        """
        If we have any changes with the new template files, make a git commit
        """
        # Check that we have something to commit
        if not self.repo.is_dirty(untracked_files=True):
            print("[bold blue]Template contains no changes - no new commit created")
            return False
        # Commit changes
        try:
            # git add only non-blacklisted files
            print("[bold blue]Staging template.")
            # add all files to stage
            self.repo.git.add(A=True)
            # get all changed/modified files during the sync (including blacklisted ones)
            changed_files = [item.a_path for item in self.repo.index.diff("HEAD")]
            blacklisted_changed_files = set()
            for pattern in self.blacklisted_globs:
                # keep track of all staged files matching a glob from the mlf_core.cfg file
                # those files will be excluded from syncing but will still be available in every new created projects
                blacklisted_changed_files |= {file for file in fnmatch.filter(changed_files, pattern)}
            nl = "\n"
            log.debug(
                f"Blacklisted (unsynced) files are:{nl}{nl.join(file for file in blacklisted_changed_files)}"
                if blacklisted_changed_files
                else "No blacklisted files for syncing found."
            )
            # each blacklisted file must be first unstaged and then get its changes discarded, so it will not be marked as modified by git
            for blacklisted_file in blacklisted_changed_files:
                log.debug(f"Unstaging {blacklisted_file}")
                Popen(["git", "reset", blacklisted_file], stdout=PIPE, stderr=PIPE, universal_newlines=True)
                log.debug(f"Discarding changes in working directory of file {blacklisted_file}")
                Popen(["git", "checkout", "--", blacklisted_file], stdout=PIPE, stderr=PIPE, universal_newlines=True)

            files_to_commit = [file for file in changed_files if file not in blacklisted_changed_files]
            log.debug(
                f"Files to commit are:{nl}{nl.join(file for file in files_to_commit)}"
                if files_to_commit
                else "No files to commit found."
            )
            if files_to_commit:
                print("[bold blue]Committing changes of non blacklisted files.")
                Popen(
                    ["git", "commit", "-m", "mlf-core sync", *files_to_commit],
                    stdout=PIPE,
                    stderr=PIPE,
                    universal_newlines=True,
                )
                print("[bold blue]Stashing and saving TEMPLATE branch changes!")
                Popen(["git", "stash"], stdout=PIPE, stderr=PIPE, universal_newlines=True)
                self.made_changes = True
                print("[bold blue]Committed changes to TEMPLATE branch")

            # if a file was added to the template, but is blacklisted, it remains untracked; so it should be removed
            for untracked_file in self.repo.untracked_files:
                log.debug(f"Removing untracked file {untracked_file}")
                Path(untracked_file).unlink()
        except Exception as e:
            print(f"[bold red]Could not commit changes to TEMPLATE:\n{e}")
            sys.exit(1)
        return True

    def push_template_branch(self):
        """
        If there are any changes to the template, push the TEMPLATE branch to the default remote
        and push to the actual sync temporary branch, where a PR is actually created from to development branch.
        """
        print(f"[bold blue]Pushing TEMPLATE branch to remote: {os.path.basename(self.project_dir)}")
        try:
            log.debug("Getting origin as remote.")
            origin = self.repo.remote("origin")
            log.debug("Setting TEMPLATE branch as upstream tracking branch.")
            self.repo.head.ref.set_tracking_branch(origin.refs.TEMPLATE)
            log.debug("Pushing to upstream branch TEMPLATE.")
            self.repo.git.push(force=True)
            print(f"[bold blue]Checking out to new branch mlf_core_sync_v{self.new_template_version}")
            log.debug(f"git checkout -b mlf_core_sync_v{self.new_template_version}")
            self.repo.git.checkout("-b", f"mlf_core_sync_v{self.new_template_version}")
            log.debug(f"git push origin mlf_core_sync_v{self.new_template_version}")
            print(f"[bold blue]Pushing to remote branch mlf_core_sync_v{self.new_template_version}")
            self.repo.remotes.origin.push(
                refspec=f"mlf_core_sync_v{self.new_template_version}:mlf_core_sync_v{self.new_template_version}"
            )
        except git.exc.GitCommandError as e:
            print(f"Could not push TEMPLATE or mlf_core_sync_v{self.new_template_version} branch:\n{e}")
            sys.exit(1)

    def make_pull_request(self):
        """
        Create a pull request to a base branch from a head branch (that is, a temporary branch only created for syncing the new template version)
        """
        log.debug("Preparing PR contents to submit a sync PR.")
        if self.dot_mlf_core["is_github_orga"]:
            self.repo_owner = self.dot_mlf_core["github_orga"]
        pr_title = f"Important mlf-core template update {self.new_template_version} released!"
        pr_body_text = (
            "A new release of the main template in mlf-core has just been released. "
            "This automated pull-request attempts to apply the relevant updates to this Project.\n\n"
            "Please make sure to merge this pull-request as soon as possible. "
            "Once complete, make a new minor release of your Project.\n\n"
            "For more information on the actual changes, read the latest mlf-core changelog."
        )
        log.debug(f"PR title is:\n{pr_title}")
        log.debug(f"PR body is:\n{pr_body_text}")

        # Check, whether a mlf-core sync PR is already open
        self.check_pull_request_exists()
        # Submit the new pull request with the latest mlf-core sync changes
        self.submit_pull_request(pr_title, pr_body_text)

    def submit_pull_request(self, pr_title: str, pr_body_text: str):
        """
        Create a new pull-request on GitHub
        """
        repo = self.github.get_repo(f'{self.repo_owner}/{self.dot_mlf_core["project_slug"]}')
        try:
            repo.create_pull(
                title=pr_title,
                body=pr_body_text,
                head=f"mlf_core_sync_v{self.new_template_version}",
                base="development",
                maintainer_can_modify=True,
            )
            print("[bold blue]Successfully created PR!")

        # Something went wrong
        except GithubException as e:
            print(f"[bold red]Could not create a PR due to the following error:\n{e}")
            sys.exit(1)

    def check_pull_request_exists(self) -> bool:
        """
        Check, whether a mlf-core sync PR is already open. If so, close this one, as it is outdated.

        :return Whether a mlf-core sync PR is already open or not
        """
        repo = self.github.get_repo(f'{self.repo_owner}/{self.dot_mlf_core["project_slug"]}')
        # query all open PRs
        log.debug("Querying open PRs to check if a sync PR already exists.")
        # iterate over the open PRs of the repo to check if a mlf-core sync PR is open
        for pull_request in repo.get_pulls(state="open"):
            log.debug("Already open sync PR has been found.")
            # if an older, outdated mlf-core sync PR is still open, close it first
            if "Important mlf-core template update" in pull_request.title:
                print(
                    "[bold blue]Detected open, but outdated mlf-core sync PR. Closing it in favor of the newest sync PR!"
                )
                pull_request.edit(state="closed")
                return True
        return False

    def should_run_sync(self) -> bool:
        """
        Check, whether sync should run. This depends on two things:
        1.) First check, whether the user disabled sync in the mlf_core.cfg file with sync_enabled = False in the [sync] section
        2.) Secondly, whether the configured level in the [sync_level] section does not restrict the sync.
        Possible levels are:
            - patch: Always create a pull request (lower bound)
            - minor: Create a pull request if it's a minor or major change
            - major: Create a pull request only if it's a major change
        :return: Whether a sync should run determined by the rules above
        """
        log.debug(f"Checking sync level constraints using parsed results from {self.project_dir}/mlf_core.cfg")
        try:
            parser = ConfigParser()
            parser.read(f"{self.project_dir}/mlf_core.cfg")
            sync_enabled = parser.items("sync")
            # sync is enabled
            if sync_enabled[0][1].lower() in {"yes", "y", "true"}:
                pass
            # sync is disabled
            elif sync_enabled[0][1].lower() in {"no", "n", "false"}:
                return False
            # misconfigured sync_enabled with some unknown value
            else:
                print(
                    f"[bold blue]Unknown value {sync_enabled[0][1]} for sync_enabled config.\nAllowed values are [bold green]Yes, yes, True, true, Y, y "
                    f"[bold blue] or [bold green]No, no, False, false, N, n [bold blue]!"
                )
                return False
            # now, check for the sync level
            level_item = list(parser.items("sync_level"))
            log.debug(f"Parsing level constraint returned: {level_item}.")
            # check for proper configuration if the sync_level section (only one item named mlf_core_sync_level with valid levels major or minor
            if (
                len(level_item) != 1
                or "sync_level" not in level_item[0][0]
                or not any(level_item[0][1] == valid_lvl for valid_lvl in ["major", "minor", "patch"])
            ):
                print(
                    "[bold red]Your mlf_core_sync_level section is missconfigured. Make sure that it only contains one item named mlf_core_sync_level "
                    "with only valid levels patch, minor or major!"
                )
                sys.exit(1)
            # check in case of minor update that level is not set to major (major case must not be handled as level is a lower bound)
            if self.patch_update:
                log.debug("Checking whether constraints allow patch updates.")
                return level_item[0][1] != "minor" and level_item[0][1] != "major"
            elif self.minor_update:
                log.debug("Checking whether constraints allow minor updates.")
                return level_item[0][1] != "major"
            else:
                log.debug("All updates are allowed because patch level is set.")
                return True
        # mlf_core.cfg file was not found or has no section sync_level
        except NoSectionError:
            print(
                "[bold red]Could not read from mlf_core.cfg file. "
                "Make sure your specified path contains a mlf_core.cfg file and has a sync_level and sync section!"
            )
            sys.exit(1)

    def get_blacklisted_sync_globs(self) -> list:
        """
        Get all blacklisted globs from the mlf_core.cfg file.
        :return: A list of all blacklisted globs for sync (file (types) that should not be included into the sync pull request)
        """
        try:
            parser = ConfigParser()
            parser.read(f"{self.project_dir}/mlf_core.cfg")
            globs = list(parser.items("sync_files_blacklisted"))
            nl = "\n"
            log.debug(f"Returning all blacklisted files globs parsed from {self.project_dir}/mlf_core.cfg.")
            log.debug(f"Blacklisted globs were {nl}{nl.join(glob[1] for glob in globs)}")
            return [glob[1] for glob in globs]

        # mlf_core.cfg file was not found or has no section called sync_files_blacklisted
        except NoSectionError:
            print(
                "[bold red]Could not read from mlf_core.cfg file. Make sure your specified path contains a mlf_core.cfg file and has a "
                "sync_files_blacklisted section!"
            )
            sys.exit(1)

    def reset_target_dir(self):
        """
        Reset the target project directory. Check out the original branch.
        """
        print(f"[bold blue]Checking out original branch: {self.original_branch}")
        try:
            self.repo.git.checkout(self.original_branch)
        except git.exc.GitCommandError as e:
            print(f"[bold red]Could not reset to original branch {self.from_branch}:\n{e}")
            sys.exit(1)

    @staticmethod
    def update_sync_token(project_name: str, gh_username: str = "") -> None:
        """
        Update the sync token secret for the repository.

        :param project_name Name of the users project
        :param gh_username The Github username (only gets passed, if the repo is an orga repo)
        """
        gh_username = (
            load_yaml_file(ConfigCommand.CONF_FILE_PATH)["github_username"] if not gh_username else gh_username
        )
        # get the personal access token for user authentification
        log.debug("Asking for updated sync token value.")
        updated_sync_token = mlf_core_questionary_or_dot_mlf_core(
            function="password", question="Please enter your updated sync token value"
        )
        print(f"[bold blue]\nUpdating sync secret for project {project_name}.")
        create_sync_secret(gh_username, project_name, updated_sync_token)  # type: ignore
        print(f"[bold blue]\nSuccessfully updated sync secret for project {project_name}.")

    @staticmethod
    def has_template_version_changed(project_dir: Path) -> Tuple[bool, bool, bool, str, str]:
        """
        Check, if the mlf-core template has been updated since last check/sync of the user.

        :return: Both false if no versions changed or a micro change happened (for ex. 1.2.3 to 1.2.4). Return is_major_update True if a major version release
        happened for the mlf-core template (for example 1.2.3 to 2.0.0). Return is_minor_update True if a minor change happened (1.2.3 to 1.3.0).
        Return is_patch_update True if its a micro update (for example 1.2.3 to 1.2.4).
        mlf-core will use this to decide which syncing strategy to apply. Also return both versions.
        """
        # Try to compare against the development branch, since it is the most up to date (usually).
        # If a development branch does not exist compare against master.
        repo = git.Repo(project_dir)
        try:
            repo.git.checkout("development")
        except git.exc.GitCommandError:  # type: ignore[misc]
            print("[bold red]Could not checkout development branch. Trying to checkout master...")
            try:
                repo.git.checkout("master")
            except git.exc.GitCommandError as e:  # type: ignore[misc]
                print(f"[bold red]Could not checkout master branch.\n{e}")
                sys.exit(1)

        log.debug("Loading the project's template version and the mlf-core template version.")
        template_version_last_sync, template_handle = TemplateSync.sync_load_project_template_version_and_handle(
            project_dir
        )
        template_version_last_sync = version.parse(template_version_last_sync)  # type: ignore
        current_mlf_core_template_version = version.parse(TemplateSync.sync_load_template_version(template_handle))
        log.debug(
            f"Projects template version is {template_version_last_sync} and mlf-core template version is {current_mlf_core_template_version}"
        )
        is_major_update, is_minor_update, is_patch_update = False, False, False

        # check if a major change happened (for example 1.2.3 to 2.0.0)
        if template_version_last_sync.major < current_mlf_core_template_version.major:  # type: ignore
            is_major_update = True
        # check if minor update happened (for example 1.2.3 to 1.3.0)
        elif template_version_last_sync.minor < current_mlf_core_template_version.minor:  # type: ignore
            is_minor_update = True
        # check if a patch update happened (for example 1.2.3 to 1.2.4)
        elif template_version_last_sync.micro < current_mlf_core_template_version.micro:  # type: ignore
            is_patch_update = True
        return (
            is_major_update,
            is_minor_update,
            is_patch_update,
            str(template_version_last_sync),
            str(current_mlf_core_template_version),
        )

    @staticmethod
    def sync_load_template_version(handle: str) -> str:
        """
        Load the version of the template available from mlf-core specified by the handler for syncing.

        :param handle: The template handle
        :return: The actual version number of the template in mlf-core
        """
        top_path = f"{os.path.dirname(__file__)}/.."
        available_templates_path = f"{str(top_path)}/create/templates/available_templates.yml"
        log.debug(
            f"Using available templates file from {available_templates_path} to load current mlf-core template version."
        )
        return load_mlf_core_template_version(handle, available_templates_path)

    @staticmethod
    def sync_load_project_template_version_and_handle(project_dir: Path) -> Tuple[str, str]:
        """
        Return the project template version since last sync for user (if no sync happened, return initial create version of the template)

        :param project_dir: Top level path to users project directory
        """
        return load_project_template_version_and_handle(project_dir)



import logging
import os
import sys
from pathlib import Path

import appdirs
from cryptography.fernet import Fernet
from mlf_core.common.levensthein_dist import most_similar_command
from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from rich import print
from rich.box import HEAVY_HEAD
from rich.console import Console
from rich.style import Style
from rich.table import Table
from ruamel.yaml import YAML

log = logging.getLogger(__name__)


class ConfigCommand:
    """
    Class for the config command. Sets general configurations such as full name, email and Github username.
    """

    # path where the config file is stored for mlf_core
    CONF_FILE_PATH = f'{appdirs.user_config_dir(appname="mlf_core")}/mlf_core_cfg.yml'
    KEY_PAT_FILE = f'{appdirs.user_config_dir(appname="mlf_core")}/.ct_keys'

    @staticmethod
    def all_settings() -> None:
        """
        Set general settings and PAT.
        """
        if not ConfigCommand.check_mlf_core_config_dir_exists():
            ConfigCommand.create_config_dirs()
        ConfigCommand.config_general_settings()
        ConfigCommand.config_pat()

    @staticmethod
    def config_general_settings() -> None:
        """
        Set full_name, email and github username for reuse in any project created further on. Defaults for prompts
        are set to the previous configured values (if any).
        """
        already_configured = False
        settings = {}
        config_path = Path(ConfigCommand.CONF_FILE_PATH)
        if config_path.exists() and "GITHUB_ACTIONS" not in os.environ:
            already_configured = True
            settings = load_yaml_file(ConfigCommand.CONF_FILE_PATH)

        full_name = mlf_core_questionary_or_dot_mlf_core(
            function="text",
            question="Full name",
            default="Homer Simpson" if not already_configured else settings["full_name"],
        )
        email = mlf_core_questionary_or_dot_mlf_core(
            function="text",
            question="Personal or work email",
            default="homer.simpson@example.com" if not already_configured else settings["email"],
        )
        github_username = mlf_core_questionary_or_dot_mlf_core(  # type: ignore
            function="text",
            question="Github username",
            default="HomerGithub" if not already_configured else settings["github_username"],
        ).lower()

        # if the configs exist, just update them
        if os.path.exists(ConfigCommand.CONF_FILE_PATH):
            log.debug(f"Configuration was found at {ConfigCommand.CONF_FILE_PATH}. Updating configuration...")
            path = Path(ConfigCommand.CONF_FILE_PATH)
            yaml = YAML()
            settings = yaml.load(path)

            # update the full_name and email
            settings["full_name"] = full_name
            settings["email"] = email
            settings["github_username"] = github_username
            yaml.dump(settings, Path(ConfigCommand.CONF_FILE_PATH))

        # the configs don´t exist -> create them
        else:
            log.debug(
                f"Configuration was not found at {ConfigCommand.CONF_FILE_PATH}. Creating a new configuration file."
            )
            settings = {"full_name": full_name, "email": email, "github_username": github_username}
            yaml = YAML()
            yaml.dump(settings, Path(ConfigCommand.CONF_FILE_PATH))

    @staticmethod
    def config_pat() -> None:
        """
        Set the personal access token (PAT) for automatic Github repo creation.
        """
        if not ConfigCommand.check_mlf_core_config_dir_exists():
            ConfigCommand.create_config_dirs()

        try:
            path = Path(ConfigCommand.CONF_FILE_PATH)
            yaml = YAML()
            settings = yaml.load(path)
            if not all(attr in settings for attr in ["full_name", "github_username", "email"]):
                print("[bold red]The mlf-core config file misses some required attributes!")
                print("[bold blue]Lets set them before setting your Github personal access token!")
                ConfigCommand.config_general_settings()

        except FileNotFoundError:
            print("[bold red]Cannot find a mlf-core config file. Is this your first time using mlf-core?")
            print("[bold blue]Lets create one before setting your Github personal access token!")
            ConfigCommand.config_general_settings()

        if mlf_core_questionary_or_dot_mlf_core(
            function="confirm",
            question="Do you want to configure your GitHub personal access token right now?\n"
            "You can still configure it later "
            "by calling    mlf-core config pat",
            default="Yes",
        ):
            print(
                "[bold blue]mlf-core requires your Github Access token to have full repository, workflow and create/update packages permissions!"
            )
            access_token = mlf_core_questionary_or_dot_mlf_core(
                function="password", question="Please enter your Github Access token"
            )
            access_token_b = access_token.encode("utf-8")  # type: ignore

            # ask for confirmation since this action will delete the PAT irrevocably if the user has not saved it anywhere else
            if not mlf_core_questionary_or_dot_mlf_core(
                function="confirm",
                question="You´re about to update your personal access token. This action cannot be undone!\n"
                "Do you really want to continue?",
                default="Yes",
            ):
                sys.exit(1)

            # encrypt the given PAT and save the encryption key and encrypted PAT in separate files
            print("[bold blue]Generating key for encryption.")
            log.debug("Generating personal access key.")
            key = Fernet.generate_key()
            fer = Fernet(key)
            log.debug("Encrypting personal access token. ")
            print("[bold blue]Encrypting personal access token.")
            encrypted_pat = fer.encrypt(access_token_b)

            # write key
            with open(ConfigCommand.KEY_PAT_FILE, "wb") as f:
                f.write(key)

            path = Path(ConfigCommand.CONF_FILE_PATH)
            yaml = YAML()
            settings = yaml.load(path)
            settings["pat"] = encrypted_pat
            log.debug(f"Dumping configuration to {ConfigCommand.CONF_FILE_PATH}")
            yaml.dump(settings, Path(ConfigCommand.CONF_FILE_PATH))

    @staticmethod
    def view_current_config() -> None:
        """
        Print the current users mlf-core configuration.
        """
        # load current settings
        try:
            log.debug(f"Fetching current cookietemple configuration at {ConfigCommand.CONF_FILE_PATH}.")
            settings = load_yaml_file(ConfigCommand.CONF_FILE_PATH)
            log.debug("Creating configuration table")
            # create the table and print
            table = Table(
                title="[bold]Your current configuration",
                title_style="blue",
                header_style=Style(color="blue", bold=True),
                box=HEAVY_HEAD,
            )
            table.add_column("Name", style="green")
            table.add_column("Value", style="green")
            # add rows to the table consisting of the name and value of the current setting
            for (name, value) in settings.items():
                # don't print token directly, just inform it's set
                if name == "pat":
                    table.add_row("[bold]Personal access token", "TOKEN_IS_SET")
                else:
                    table.add_row(f'[bold]{name.capitalize().replace("_", " ")}', f"[white]{value}")
            # don't print PAT directly but inform if not set
            if "pat" not in settings.keys():
                table.add_row("[bold]Personal access token", "[red]NO_TOKEN_SET")

            console = Console()
            console.print(table)

        except (FileNotFoundError, KeyError):
            print(
                "[bold red]Did not found a mlf-core config file!\nIf this is your first time running mlf-core you can set them using mlf-core "
                "config general"
            )

    @staticmethod
    def similar_handle(section: str) -> None:
        """
        Try to use/suggest a similar handle if user misspelled it.

        :param section: The handle inputted by the user.
        """
        com_list, action = most_similar_command(section.lower(), {"general", "pat", "all"})
        # use best match
        if len(com_list) == 1 and action == "use":
            print(f"[bold blue]Unknown handle {section}. Will use best match {com_list[0]}.\n")
            ConfigCommand.handle_switcher().get(com_list[0], lambda: "Invalid handle!")()
        # suggest best match
        elif len(com_list) == 1 and action == "suggest":
            print(f"[bold blue]Unknown handle {section}. Did you mean {com_list[0]}?")
            sys.exit(1)
            # multiple best matches
        elif len(com_list) > 1:
            nl = "\n"
            print(
                f"[bold red]Unknown handle '{section}'.\nMost similar handles are: [green]{nl}{nl.join(sorted(com_list))}"
            )
        else:
            # unknown handle and no best match found
            print("[bold red]Unknown handle. [green]See mlf-core --help [red]for more information on valid handles")

    @staticmethod
    def check_mlf_core_config_dir_exists() -> bool:
        """
        Check whether the config directory for mlf-core exists.
        """
        log.debug(f"Checking whether a config directory already exists at {Path(ConfigCommand.CONF_FILE_PATH).parent}.")
        if not os.path.exists(Path(ConfigCommand.CONF_FILE_PATH).parent):
            log.debug("Config directory did not exist. Creating it.")
            return False
        return True

    @staticmethod
    def create_config_dirs() -> None:
        """
        Create the config directory, if none exists (and parent directories, if necessary)
        """
        os.makedirs(Path(ConfigCommand.CONF_FILE_PATH).parent)

    @staticmethod
    def handle_switcher() -> dict:
        """
        Just a helper to switch between handles.

        :return: The switcher with all handles.
        """
        switcher = {
            "all": ConfigCommand.all_settings,
            "general": ConfigCommand.config_general_settings,
            "pat": ConfigCommand.config_pat,
        }
        return switcher



import time
from argparse import ArgumentParser

import GPUtil
import mlflow
import mlflow.xgboost
import xgboost as xgb
from data_loading.data_loader import load_train_test_data
from rich import print, traceback

from mlf_core.mlf_core import MLFCore


def start_training():
    parser = ArgumentParser(description='XGBoost Example')
    parser.add_argument(
        '--max_epochs',
        type=int,
        default=25,
        help='Number of epochs to train',
    )
    parser.add_argument(
        '--general-seed',
        type=int,
        default=0,
        help='General Python, Python random and Numpy seed.',
    )
    parser.add_argument(
        '--xgboost-seed',
        type=int,
        default=0,
        help='XGBoost specific random seed.',
    )
    parser.add_argument(
        '--cuda',
        type=bool,
        default=True,
        help='Enable or disable CUDA support.',
    )
    parser.add_argument(
        '--single-precision-histogram',
        type=bool,
        default=True,
        help='Enable or disable single precision histogram calculation.',
    )
    avail_gpus = GPUtil.getGPUs()
    args = parser.parse_args()
    dict_args = vars(args)
    use_cuda = True if dict_args['cuda'] and len(avail_gpus) > 0 else False
    if use_cuda:
        print(f'[bold blue]Using {len(avail_gpus)} GPUs!')
    else:
        print('[bold blue]No GPUs detected. Running on the CPU')

    with mlflow.start_run():
        # Enable the logging of all parameters, metrics and models to mlflow
        mlflow.autolog(1)

        # Log hardware and software
        MLFCore.log_sys_intel_conda_env()

        # Fetch and prepare data
        dtrain, dtest = load_train_test_data()

        # TODO MLF-CORE: Enable input data logging
        # MLFCore.log_input_data('data/')

        # Set XGBoost parameters
        param = {'objective': 'multi:softmax',
                 'num_class': 8,
                 'single_precision_histogram': True if dict_args['single_precision_histogram'] == 'True' else False,
                 'subsample': 0.5,
                 'colsample_bytree': 0.5,
                 'colsample_bylevel': 0.5}

        # Set random seeds
        MLFCore.set_general_random_seeds(dict_args["general_seed"])
        MLFCore.set_xgboost_random_seeds(dict_args["xgboost_seed"], param)

        # Set CPU or GPU as training device
        if use_cuda:
            param['tree_method'] = 'gpu_hist'
        else:
            param['tree_method'] = 'hist'

        # Train on the chosen device
        results = {}
        runtime = time.time()
        xgb.train(param, dtrain, dict_args['max_epochs'], evals=[(dtest, 'test')], evals_result=results)
        device = 'GPU' if use_cuda else 'CPU'
        if use_cuda:
            print(f'[bold green]{device} Run Time: {str(time.time() - runtime)} seconds')


if __name__ == '__main__':
    traceback.install()
    start_training()



import pytorch_lightning as pl
from torch.utils.data import DataLoader
from torchvision import datasets, transforms


class MNISTDataModule(pl.LightningDataModule):
    def __init__(self, **kwargs):
        """
        Initialization of the data module with a train and test dataset, as well as a loader for each.
        The dataset is th example MNIST dataset
        """
        super(MNISTDataModule, self).__init__()
        self.df_train = None
        self.df_test = None
        self.train_data_loader = None
        self.test_data_loader = None
        self.args = kwargs

        # transforms for images
        self.transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))])

    def setup(self, stage=None):
        """
        Downloads the data, parse it and split the data into train, test, validation data
        :param stage: Stage - training or testing
        """
        self.df_train = datasets.MNIST('dataset', download=True, train=True, transform=self.transform)
        self.df_test = datasets.MNIST('dataset', download=True, train=False, transform=self.transform)

    def train_dataloader(self):
        """
        :return: output - Train data loader for the given input
        """
        return DataLoader(self.df_train, batch_size=self.args['training_batch_size'], num_workers=self.args['num_workers'], shuffle=True)

    def test_dataloader(self):
        """
        :return: output - Test data loader for the given input
        """
        return DataLoader(self.df_test, batch_size=self.args['test_batch_size'], num_workers=self.args['num_workers'], shuffle=False)



import xgboost as xgb
from sklearn.datasets import fetch_covtype
from sklearn.model_selection import train_test_split


def load_train_test_data():
    # Fetch dataset using sklearn
    dataset = fetch_covtype()
    X = dataset.data
    y = dataset.target

    # Create 0.75/0.25 train/test split
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25, train_size=0.75, random_state=0)

    # Convert input data from numpy to XGBoost format
    dtrain = xgb.DMatrix(X_train, label=y_train)
    dtest = xgb.DMatrix(X_test, label=y_test)

    return dtrain, dtest



#!/usr/bin/env python

"""The setup script."""


import os

import {{ cookiecutter.project_slug_no_hyphen }} as module
from setuptools import find_packages, setup


def walker(base, *paths):
    file_list = set([])
    cur_dir = os.path.abspath(os.curdir)

    os.chdir(base)
    try:
        for path in paths:
            for dname, dirs, files in os.walk(path):
                for f in files:
                    file_list.add(os.path.join(dname, f))
    finally:
        os.chdir(cur_dir)

    return list(file_list)


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup_requirements = 'pytest-runner'

test_requirements = 'pytest'

{%- set license_classifiers = {
    'MIT': 'License :: OSI Approved :: MIT License',
    'BSD': 'License :: OSI Approved :: BSD License',
    'ISC': 'License :: OSI Approved :: ISC License (ISCL)',
    'Apache2.0': 'License :: OSI Approved :: Apache Software License',
    'GNUv3': 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
} %}

setup(
    author="{{ cookiecutter.full_name.replace('\"', '\\\"') }}",
    author_email='{{ cookiecutter.email }}',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
{%- if cookiecutter.license in license_classifiers %}
        '{{ license_classifiers[cookiecutter.license] }}',
{%- endif %}
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="{{ cookiecutter.project_short_description }}",
    entry_points={
        'console_scripts': [
            '{{ cookiecutter.project_name }}={{ cookiecutter.project_slug|replace('-', '_') }}.cli_{% if cookiecutter.framework == 'pytorch' -%}pytorch{% elif cookiecutter.framework == 'tensorflow' -%}tensorflow{% elif cookiecutter.framework == 'xgboost' -%}xgboost{% endif %}:main',
        ],
    },
    install_requires=requirements,
{%- if cookiecutter.license in license_classifiers %}
    license="{{ cookiecutter.license }}",
{%- endif %}
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='{{ cookiecutter.project_slug }}',
    name='{{ cookiecutter.project_name }}',
    packages=find_packages(include=['{{ cookiecutter.project_slug_no_hyphen }}', '{{ cookiecutter.project_slug_no_hyphen }}.*']),
    package_data={
        module.__name__: walker(
            os.path.dirname(module.__file__),
            'models', 'data'
        ),
    },
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    url='https://github.com/{{ cookiecutter.github_username }}/{{ cookiecutter.project_slug }}',
    version='{{ cookiecutter.version }}',
    zip_safe=False,
)



import logging
import os
import re
import shutil
import sys
import tempfile
from collections import OrderedDict
from dataclasses import asdict
from distutils.dir_util import copy_tree
from pathlib import Path

import requests
from cookiecutter.main import cookiecutter
from rich import print
from ruamel.yaml import YAML

import mlf_core
from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.config.config import ConfigCommand
from mlf_core.create.domains.mlf_core_template_struct import MlfcoreTemplateStruct
from mlf_core.create.github_support import create_push_github_repository, is_git_repo, load_github_username
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from mlf_core.lint.lint import lint_project
from mlf_core.util.dir_util import delete_dir_tree
from mlf_core.util.docs_util import fix_short_title_underline

log = logging.getLogger(__name__)


class TemplateCreator:
    """
    The base class for all creators.
    It holds the basic template information that are common across all templates (like a project name).
    Furthermore it defines methods that are basic for the template creation process.
    """

    def __init__(self, creator_ctx: MlfcoreTemplateStruct):
        self.WD = os.path.dirname(__file__)
        self.TEMPLATES_PATH = f"{self.WD}/templates"
        self.COMMON_FILES_PATH = f"{self.TEMPLATES_PATH}/common_all_files"
        self.COMMON_MLFLOW_FILES_PATH = f"{self.TEMPLATES_PATH}/common_mlflow_files"
        self.AVAILABLE_TEMPLATES_PATH = f"{self.TEMPLATES_PATH}/available_templates.yml"
        self.AVAILABLE_TEMPLATES = load_yaml_file(self.AVAILABLE_TEMPLATES_PATH)
        self.CWD = Path.cwd()
        self.creator_ctx = creator_ctx

    def process_common_operations(
        self,
        path: Path,
        skip_fix_underline=False,
        domain: str = None,
        subdomain: str = None,
        language: str = None,
        dot_mlf_core: OrderedDict = None,
    ) -> None:
        """
        Create all stuff that is common for mlf-core's template creation process; in detail those things are:
        create and copy common files, fix docs style, lint the project and ask whether the user wants to create a github repo.
        """
        # create the common files and copy them into the templates directory
        self.create_common_files(domain="all", common_files_path=self.COMMON_FILES_PATH)
        # key in the switcher indicates, whether there are domain specific files or not (None)
        domain_switcher = {"mlflow": "mlflow", "package": None}
        # if project is a project with domain specific files, copy all common files for the domain of this project
        try:
            domain_specific_files = domain_switcher[self.creator_ctx.domain]
        # this should only be the case, if mlf-core is developed further and new domains are added, therefore the error message
        except KeyError:
            print(
                f"[bold red]Unknown domain {self.creator_ctx.domain}! This domain seems to be new and must be added into the domain switcher!"
            )
            sys.exit(1)
        if domain_specific_files:
            self.create_common_files(domain=domain_specific_files, common_files_path=self.COMMON_MLFLOW_FILES_PATH)

        self.create_dot_mlf_core(template_version=self.creator_ctx.template_version)

        project_path = f"{self.CWD}/{self.creator_ctx.project_slug}"

        # Ensure that docs are looking good (skip if flag is set)
        if not skip_fix_underline:
            fix_short_title_underline(f"{project_path}/docs/index.rst")

        # Lint the project to verify that the new template adheres to all standards
        lint_project(project_path)

        if self.creator_ctx.is_github_repo and not dot_mlf_core:
            # rename the currently created template to a temporary name, create Github repo, push, remove temporary template
            tmp_project_path = f"{project_path}_mlf_core_tmp"
            os.mkdir(tmp_project_path)
            create_push_github_repository(project_path, self.creator_ctx, tmp_project_path)
            shutil.rmtree(tmp_project_path, ignore_errors=True)

        if subdomain:
            print(
                f"\n[bold blue]Please visit: https://mlf-core.readthedocs.io/en/latest/available_templates/available_templates.html"
                f"#{domain}-{subdomain}-{language} for more information about how to use your chosen template."
            )
        else:
            print(
                f"\n[bold blue]Please visit: https://mlf-core.readthedocs.io/en/latest/available_templates/available_templates.html"
                f"#{domain}-{language} for more information about how to use your chosen template."
            )

        # do not move if path is current working directory or a directory named like the project in the current working directory (second is default case)
        if path != self.CWD and path != Path(self.CWD / self.creator_ctx.project_slug_no_hyphen):
            shutil.move(
                f"{self.CWD}/{self.creator_ctx.project_slug_no_hyphen}",
                f"{path}/{self.creator_ctx.project_slug_no_hyphen}",
            )

    def create_template_without_subdomain(self, domain_path: str) -> None:
        """
        Creates a chosen template that does **not** have a subdomain.
        Calls cookiecutter on the main chosen template.

        :param domain_path: Path to the template, which is still in cookiecutter format
        """
        # Target directory is already occupied -> overwrite?
        occupied = os.path.isdir(f"{self.CWD}/{self.creator_ctx.project_slug}")
        if occupied:
            self.directory_exists_warning()

            # Confirm proceeding with overwriting existing directory
            if mlf_core_questionary_or_dot_mlf_core("confirm", "Do you really want to continue?", default="Yes"):
                cookiecutter(
                    f"{domain_path}/{self.creator_ctx.domain}_{self.creator_ctx.language.lower()}",
                    no_input=True,
                    overwrite_if_exists=True,
                    extra_context=self.creator_ctx_to_dict(),
                )
            else:
                print("[bold red]Aborted! Canceled template creation!")
                sys.exit(0)
        else:
            cookiecutter(
                f"{domain_path}/{self.creator_ctx.domain}_{self.creator_ctx.language.lower()}",
                no_input=True,
                overwrite_if_exists=True,
                extra_context=self.creator_ctx_to_dict(),
            )

    def create_template_with_subdomain(self, domain_path: str, subdomain: str) -> None:
        """
        Creates a chosen template that **does** have a subdomain.
        Calls cookiecutter on the main chosen template.

        :param domain_path: Path to the template, which is still in cookiecutter format
        :param subdomain: Subdomain of the chosen template
        """
        occupied = os.path.isdir(f"{self.CWD}/{self.creator_ctx.project_slug}")
        if occupied:
            self.directory_exists_warning()

            # Confirm proceeding with overwriting existing directory
            if mlf_core_questionary_or_dot_mlf_core("confirm", "Do you really want to continue?", default="Yes"):
                delete_dir_tree(Path(f"{self.CWD}/{self.creator_ctx.project_slug}"))
                cookiecutter(
                    f"{domain_path}/{subdomain}_{self.creator_ctx.language.lower()}",
                    no_input=True,
                    overwrite_if_exists=True,
                    extra_context=self.creator_ctx_to_dict(),
                )

            else:
                print("[bold red]Aborted! Canceled template creation!")
                sys.exit(0)
        else:
            cookiecutter(
                f"{domain_path}/{subdomain}_{self.creator_ctx.language.lower()}",
                no_input=True,
                overwrite_if_exists=True,
                extra_context=self.creator_ctx_to_dict(),
            )

    def create_template_with_subdomain_framework(self, domain_path: str, subdomain: str, framework: str) -> None:
        """
        Creates a chosen template that **does** have a subdomain.
        Calls cookiecutter on the main chosen template.

        :param domain_path: Path to the template, which is still in cookiecutter format
        :param subdomain: Subdomain of the chosen template
        :param framework: Chosen framework
        """
        occupied = os.path.isdir(f"{self.CWD}/{self.creator_ctx.project_slug}")
        if occupied:
            self.directory_exists_warning()

            # Confirm proceeding with overwriting existing directory
            if mlf_core_questionary_or_dot_mlf_core("confirm", "Do you really want to continue?", default="Yes"):
                cookiecutter(
                    f"{domain_path}/{subdomain}_{self.creator_ctx.language.lower()}/{framework}",
                    no_input=True,
                    overwrite_if_exists=True,
                    extra_context=self.creator_ctx_to_dict(),
                )

            else:
                print("[bold red]Aborted! Canceled template creation!")
                sys.exit(0)
        else:
            cookiecutter(
                f"{domain_path}/{subdomain}_{self.creator_ctx.language.lower()}/{framework}",
                no_input=True,
                overwrite_if_exists=True,
                extra_context=self.creator_ctx_to_dict(),
            )

    def prompt_general_template_configuration(self, dot_mlf_core: OrderedDict):
        """
        Prompts the user for general options that are required by all templates.
        Options are saved in the creator context manager object.
        """
        try:
            if dot_mlf_core:
                # try to read name and email from existing config file
                self.creator_ctx.full_name = dot_mlf_core["full_name"]
                self.creator_ctx.email = dot_mlf_core["email"]
            else:
                self.creator_ctx.full_name = load_yaml_file(ConfigCommand.CONF_FILE_PATH)["full_name"]
                self.creator_ctx.email = load_yaml_file(ConfigCommand.CONF_FILE_PATH)["email"]
        except FileNotFoundError:
            # style and automatic use config
            print("[bold red]Cannot find a mlf_core config file. Is this your first time using mlf-core?")
            # inform the user and config all settings (with PAT optional)
            print("[bold blue]Lets set your name, email and Github username and you´re ready to go!")
            ConfigCommand.all_settings()
            # load mail and full name
            path = Path(ConfigCommand.CONF_FILE_PATH)
            yaml = YAML(typ="safe")
            settings = yaml.load(path)
            # set full name and mail
            if dot_mlf_core:
                self.creator_ctx.full_name = dot_mlf_core["full_name"]
                self.creator_ctx.email = dot_mlf_core["email"]
            else:
                self.creator_ctx.full_name = settings["full_name"]
                self.creator_ctx.email = settings["email"]

        self.creator_ctx.project_name = mlf_core_questionary_or_dot_mlf_core(  # type: ignore
            function="text",
            question="Project name",
            default="Exploding Springfield",
            dot_mlf_core=dot_mlf_core,
            to_get_property="project_name",
        ).lower()

        # check if the project name is already taken on readthedocs.io
        # lower the string, since mlflow doesn't play with uppercase docker container names
        while self.readthedocs_slug_already_exists(self.creator_ctx.project_name) and not dot_mlf_core:
            print(f"[bold red]A project named {self.creator_ctx.project_name} already exists at readthedocs.io!")
            if mlf_core_questionary_or_dot_mlf_core(
                function="confirm",
                question="Do you want to choose another name for your project?\n"
                "Otherwise you will not be able to host your docs at readthedocs.io!",
                default="Yes",
            ):
                self.creator_ctx.project_name = mlf_core_questionary_or_dot_mlf_core(  # type: ignore
                    function="text", question="Project name", default="Exploding Springfield"
                ).lower()
            # break if the project should be named anyways
            else:
                break
        self.creator_ctx.project_slug = self.creator_ctx.project_name.replace(" ", "_")
        self.creator_ctx.project_slug_no_hyphen = self.creator_ctx.project_slug.replace("-", "_")
        self.creator_ctx.project_short_description = mlf_core_questionary_or_dot_mlf_core(  # type: ignore
            function="text",
            question="Short description of your project",
            default=f"{self.creator_ctx.project_name}" f". A mlf-core based .",
            dot_mlf_core=dot_mlf_core,
            to_get_property="project_short_description",
        )
        poss_vers = mlf_core_questionary_or_dot_mlf_core(
            function="text",
            question="Initial version of your project",
            default="0.1.0-SNAPSHOT",
            dot_mlf_core=dot_mlf_core,
            to_get_property="version",
        )

        # make sure that the version has the right format
        while not re.match(r"(?<!.)\d+(?:\.\d+){2}(?:-SNAPSHOT)?(?!.)", poss_vers) and not dot_mlf_core:  # type: ignore
            print(
                "[bold red]The version number entered does not match semantic versioning.\n"
                + "Please enter the version in the format \[number].\[number].\[number](-SNAPSHOT)!"
            )  # noqa: W605
            poss_vers = mlf_core_questionary_or_dot_mlf_core(
                function="text", question="Initial version of your project", default="0.1.0-SNAPSHOT"
            )
        self.creator_ctx.version = poss_vers  # type: ignore

        self.creator_ctx.license = mlf_core_questionary_or_dot_mlf_core(  # type: ignore
            function="select",
            question="License",
            choices=[
                "MIT",
                "BSD",
                "ISC",
                "Apache2.0",
                "GNUv3",
                "Boost",
                "Affero",
                "CC0",
                "CCBY",
                "CCBYSA",
                "Eclipse",
                "WTFPL",
                "unlicence",
                "Not open source",
            ],
            default="MIT",
            dot_mlf_core=dot_mlf_core,
            to_get_property="license",
        )

        if dot_mlf_core:
            self.creator_ctx.github_username = dot_mlf_core["github_username"]
            self.creator_ctx.creator_github_username = dot_mlf_core["creator_github_username"]
        else:
            self.creator_ctx.github_username = load_github_username()
            self.creator_ctx.creator_github_username = self.creator_ctx.github_username

    def create_common_files(self, domain: str, common_files_path: str) -> None:
        """
        Create a temporary directory for common files of a specified domain or all templates and apply cookiecutter on them.
        They are subsequently moved into the directory of the created template.
        """
        log.debug("Creating common files.")
        dirpath = tempfile.mkdtemp()
        copy_tree(common_files_path, dirpath)
        cwd_project = self.CWD
        os.chdir(dirpath)
        log.debug(f"Cookiecuttering common files at {dirpath}")
        cookiecutter(
            dirpath,
            extra_context={
                "full_name": self.creator_ctx.full_name,
                "email": self.creator_ctx.email,
                "language": self.creator_ctx.language,
                "domain": self.creator_ctx.domain,
                "project_name": self.creator_ctx.project_name,
                "project_slug": self.creator_ctx.project_slug,
                "project_slug_no_hyphen": self.creator_ctx.project_slug_no_hyphen,
                "version": self.creator_ctx.version,
                "license": self.creator_ctx.license,
                "project_short_description": self.creator_ctx.project_short_description,
                "github_username": self.creator_ctx.github_username,
                "creator_github_username": self.creator_ctx.creator_github_username,
                "framework": self.creator_ctx.language.capitalize() if domain == "mlflow" else "",
            },
            no_input=True,
            overwrite_if_exists=True,
        )

        # copy the common files directory content to the created project
        log.debug("Copying common files into the created project")
        copy_tree(f"{Path.cwd()}/common_{domain}_files", f"{cwd_project}/{self.creator_ctx.project_slug}")
        # delete the tmp cookiecuttered common files directory
        log.debug("Delete common files directory.")
        delete_dir_tree(Path(f"{Path.cwd()}/common_{domain}_files"))
        shutil.rmtree(dirpath)
        # change to recent cwd so lint can run properly
        os.chdir(str(cwd_project))

    def readthedocs_slug_already_exists(self, project_name: str) -> bool:
        """
        Test whether there´s already a project with the same name on readthedocs
        :param project_name Name of the project the user wants to create
        """
        print(f"[bold blue]Looking up {project_name} at readthedocs.io!")
        try:
            request = requests.get(f'https://{project_name.replace(" ", "")}.readthedocs.io')
            if request.status_code == 200:
                return True
        # catch exceptions when server may be unavailable or the request timed out
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
            print(
                "[bold red]Cannot check whether name already taken on readthedocs.io because its unreachable at the moment!"
            )
            return False

        return False

    def directory_exists_warning(self) -> None:
        """
        If the directory is already a git directory within the same project, print error message and exit.
        Otherwise print a warning that a directory already exists and any further action on the directory will overwrite its contents.
        """
        if is_git_repo(Path(f"{self.CWD}/{self.creator_ctx.project_slug}")):
            print(
                f"[bold red]Error: A git project named {self.creator_ctx.project_slug} already exists at [green]{self.CWD}\n"
            )
            print("[bold red]Aborting!")
            sys.exit(1)
        else:
            print(
                f"[bold yellow]WARNING: [red]A directory named {self.creator_ctx.project_slug} already exists at [blue]{self.CWD}\n"
            )
            print("Proceeding now will overwrite this directory and its content!")

    def create_dot_mlf_core(self, template_version: str):
        """
        Overrides the version with the version of the template.
        Dumps the configuration for the template generation into a .mlf_core yaml file.

        :param template_version: Version of the specific template
        """
        log.debug("Creating .mlf_core.yml file.")
        self.creator_ctx.template_version = f"{template_version} # <<MLF-CORE_NO_BUMP>>"
        self.creator_ctx.mlf_core_version = f"{mlf_core.__version__} # <<MLF-CORE_NO_BUMP>>"
        with open(f"{self.creator_ctx.project_slug}/.mlf_core.yml", "w") as f:
            yaml = YAML()
            struct_to_dict = self.creator_ctx_to_dict()
            yaml.dump(struct_to_dict, f)

    def creator_ctx_to_dict(self) -> dict:
        """
        Create a dict from the our Template Structure dataclass
        :return: The dict containing all key-value pairs with non empty values
        """
        return {key: val for key, val in asdict(self.creator_ctx).items() if val != ""}



import os
import sys

import click
import numpy as np
import torch
from rich import print, traceback

WD = os.path.dirname(__file__)


@click.command()
@click.option('-i', '--input', required=True, type=str, help='Path to data file to predict.')
@click.option('-m', '--model', type=str, help='Path to an already trained XGBoost model. If not passed a default model will be loaded.')
@click.option('-c/-nc', '--cuda/--no-cuda', type=bool, default=False, help='Whether to enable cuda or not')
@click.option('-o', '--output', type=str, help='Path to write the output to')
def main(input: str, model: str, cuda: bool, output: str):
    """Command-line interface for {{ cookiecutter.project_name }}"""

    print(r"""[bold blue]
        {{ cookiecutter.project_name }}
        """)

    print('[bold blue]Run [green]{{ cookiecutter.project_name }} --help [blue]for an overview of all commands\n')
    if not model:
        model = get_pytorch_model(f'{WD}/models/pytorch_test_model')
    else:
        model = get_pytorch_model(model)
    if cuda:
        model.cuda()
    print('[bold blue] Parsing data')
    data_to_predict = read_data_to_predict(input)
    print('[bold blue] Performing predictions')
    predictions = np.round(model.predict(data_to_predict))
    print(predictions)
    if output:
        print(f'[bold blue]Writing predictions to {output}')
        write_results(predictions, output)


def read_data_to_predict(path_to_data_to_predict: str):
    """
    Parses the data to predict and returns a full Dataset include the DMatrix
    :param path_to_data_to_predict: Path to the data on which predictions should be performed on
    """
    return


def write_results(predictions: np.ndarray, path_to_write_to) -> None:
    """
    Writes the predictions into a human readable file.
    :param predictions: Predictions as a numpy array
    :param path_to_write_to: Path to write the predictions to
    """
    pass


def get_pytorch_model(path_to_pytorch_model: str):
    """
    Fetches the model of choice and creates a booster from it.
    :param path_to_pytorch_model: Path to the xgboost model1
    """
    model = torch.load(path_to_pytorch_model)
    return model


if __name__ == "__main__":
    traceback.install()
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mlf_core',
 'mlf_core.bump_version',
 'mlf_core.common',
 'mlf_core.config',
 'mlf_core.create',
 'mlf_core.create.domains',
 'mlf_core.create.templates.common_all_files.{{cookiecutter.common_all}}.docs',
 'mlf_core.create.templates.common_mlflow_files.{{ cookiecutter.common_mlflow '
 '}}.{{ cookiecutter.project_slug_no_hyphen }}.mlf_core',
 'mlf_core.create.templates.mlflow.mlflow_pytorch.{{ cookiecutter.project_slug '
 '}}.{{ cookiecutter.project_slug_no_hyphen }}',
 'mlf_core.create.templates.mlflow.mlflow_pytorch.{{ cookiecutter.project_slug '
 '}}.{{ cookiecutter.project_slug_no_hyphen }}.data_loading',
 'mlf_core.create.templates.mlflow.mlflow_pytorch.{{ cookiecutter.project_slug '
 '}}.{{ cookiecutter.project_slug_no_hyphen }}.model',
 'mlf_core.create.templates.mlflow.mlflow_tensorflow.{{ '
 'cookiecutter.project_slug }}.{{ cookiecutter.project_slug_no_hyphen }}',
 'mlf_core.create.templates.mlflow.mlflow_tensorflow.{{ '
 'cookiecutter.project_slug }}.{{ cookiecutter.project_slug_no_hyphen '
 '}}.data_loading',
 'mlf_core.create.templates.mlflow.mlflow_tensorflow.{{ '
 'cookiecutter.project_slug }}.{{ cookiecutter.project_slug_no_hyphen }}.model',
 'mlf_core.create.templates.mlflow.mlflow_tensorflow.{{ '
 'cookiecutter.project_slug }}.{{ cookiecutter.project_slug_no_hyphen '
 '}}.training',
 'mlf_core.create.templates.mlflow.mlflow_xgboost.{{ cookiecutter.project_slug '
 '}}.{{ cookiecutter.project_slug_no_hyphen }}',
 'mlf_core.create.templates.mlflow.mlflow_xgboost.{{ cookiecutter.project_slug '
 '}}.{{ cookiecutter.project_slug_no_hyphen }}.data_loading',
 'mlf_core.create.templates.package.package_prediction.hooks',
 'mlf_core.create.templates.package.package_prediction.{{ '
 'cookiecutter.project_slug_no_hyphen }}',
 'mlf_core.create.templates.package.package_prediction.{{ '
 'cookiecutter.project_slug_no_hyphen '
 '}}.{{cookiecutter.project_slug_no_hyphen}}',
 'mlf_core.custom_cli',
 'mlf_core.info',
 'mlf_core.lint',
 'mlf_core.lint.domains',
 'mlf_core.list',
 'mlf_core.sync',
 'mlf_core.upgrade',
 'mlf_core.util']

package_data = \
{'': ['*'],
 'mlf_core.create': ['templates/*',
                     'templates/common_all_files/*',
                     'templates/common_all_files/{{cookiecutter.common_all}}/*',
                     'templates/common_all_files/{{cookiecutter.common_all}}/.github/*',
                     'templates/common_all_files/{{cookiecutter.common_all}}/.github/ISSUE_TEMPLATE/*',
                     'templates/common_all_files/{{cookiecutter.common_all}}/.github/workflows/*',
                     'templates/common_mlflow_files/*',
                     'templates/common_mlflow_files/{{ '
                     'cookiecutter.common_mlflow }}/*',
                     'templates/common_mlflow_files/{{ '
                     'cookiecutter.common_mlflow }}/.github/workflows/*',
                     'templates/mlflow/mlflow_pytorch/*',
                     'templates/mlflow/mlflow_pytorch/{{ '
                     'cookiecutter.project_slug }}/*',
                     'templates/mlflow/mlflow_pytorch/{{ '
                     'cookiecutter.project_slug }}/docs/*',
                     'templates/mlflow/mlflow_tensorflow/*',
                     'templates/mlflow/mlflow_tensorflow/{{ '
                     'cookiecutter.project_slug }}/*',
                     'templates/mlflow/mlflow_tensorflow/{{ '
                     'cookiecutter.project_slug }}/docs/*',
                     'templates/mlflow/mlflow_xgboost/*',
                     'templates/mlflow/mlflow_xgboost/{{ '
                     'cookiecutter.project_slug }}/*',
                     'templates/mlflow/mlflow_xgboost/{{ '
                     'cookiecutter.project_slug }}/docs/*',
                     'templates/package/package_prediction/*'],
 'mlf_core.create.templates.common_all_files.{{cookiecutter.common_all}}.docs': ['_static/*'],
 'mlf_core.create.templates.package.package_prediction.{{ cookiecutter.project_slug_no_hyphen }}': ['.github/*',
                                                                                                    '.github/workflows/*',
                                                                                                    'docs/*',
                                                                                                    'makefiles/*'],
 'mlf_core.create.templates.package.package_prediction.{{ cookiecutter.project_slug_no_hyphen }}.{{cookiecutter.project_slug_no_hyphen}}': ['data/*',
                                                                                                                                            'models/*']}

install_requires = \
['GitPython>=3.1.15,<4.0.0',
 'Jinja2>=2.11.3,<4.0.0',
 'PyGithub>=1.54.1,<2.0.0',
 'PyNaCl>=1.4.0,<2.0.0',
 'PyYAML>=5.4.1,<7.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'autopep8>=1.5.6,<2.0.0',
 'cffi>=1.14.5,<2.0.0',
 'click>=7.1.2,<9.0.0',
 'cookiecutter>=1.7.2,<2.0.0',
 'cryptography>=3.4.7,<36.0.0',
 'mlflow>=1.15.0,<2.0.0',
 'packaging>=20.9,<22.0',
 'questionary>=1.9.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'rich>=10.1.0,<11.0.0',
 'ruamel.yaml>=0.17.4,<0.18.0']

entry_points = \
{'console_scripts': ['mlf-core = mlf_core.__main__:main']}

setup_kwargs = {
    'name': 'mlf-core',
    'version': '1.11.4',
    'description': 'Reproducible machine learning pipelines using mlflow.',
    'long_description': ".. image:: https://user-images.githubusercontent.com/21954664/84388841-84b4cc80-abf5-11ea-83f3-b8ce8de36e25.png\n    :target: https://mlf-core.com\n    :alt: mlf-core logo\n\n|\n\n========\nmlf-core\n========\n\n|PyPI| |Python Version| |License| |Read the Docs| |Build| |Tests| |Codecov| |pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/mlf-core.svg\n   :target: https://pypi.org/project/mlf-core/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/mlf-core\n   :target: https://pypi.org/project/mlf-core\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/github/license/mlf-core/mlf-core\n   :target: https://opensource.org/licenses/Apache-2.0\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/mlf-core/latest.svg?label=Read%20the%20Docs\n   :target: https://mlf-core.readthedocs.io/\n   :alt: Read the documentation at https://mlf-core.readthedocs.io/\n.. |Build| image:: https://github.com/mlf-core/mlf-core/workflows/Build%20mlf-core%20Package/badge.svg\n   :target: https://github.com/mlf-core/mlf-core/actions?workflow=Package\n   :alt: Build Package Status\n.. |Tests| image:: https://github.com/mlf-core/mlf-core/workflows/Run%20mlf-core%20Tests/badge.svg\n   :target: https://github.com/mlf-core/mlf-core/actions?workflow=Tests\n   :alt: Run Tests Status\n.. |Codecov| image:: https://codecov.io/gh/mlf-core/mlf-core/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/mlf-core/mlf-core\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n.. image:: https://static.pepy.tech/personalized-badge/mlf-core?units=international_system&left_color=grey&right_color=green&left_text=Downloads\n   :target: https://pepy.tech/project/mlf-core\n   :alt: Pepy Downloads\n\n.. image:: https://img.shields.io/discord/742367395196305489?color=passing\n   :target: https://discord.gg/Mv8sAcq\n   :alt: Discord\n\nPreprint\n--------\n\n`mlf-core: a framework for deterministic machine learning <https://arxiv.org/abs/2104.07651>`_\n\n\nOverview\n--------\n\n.. figure:: https://user-images.githubusercontent.com/31141763/110704981-02921c80-81f6-11eb-8775-bd73f565568c.png\n   :alt: mlf-core overview\n\n   mlf-core provides CPU and GPU deterministic machine learning templates based on MLflow, Conda, Docker and a strong Github integration.\n   Templates are available for PyTorch, TensorFlow and XGBoost.\n   A custom linter ensures that projects stay deterministic in all phases of development and deployment.\n\nInstalling\n---------------\n\nStart your journey with mlf-core by installing it via ``$ pip install mlf-core``.\n\nSee `Installation  <https://mlf_core.readthedocs.io/en/latest/readme.html#installing>`_.\n\nrun\n----\nSee a mlf-core project in action.\n\n.. figure:: https://user-images.githubusercontent.com/31141763/117714817-c409e580-b1d7-11eb-9991-cb6eb58efbb7.gif\n\n\nconfig\n------\nConfigure mlf-core to get started.\n\n.. figure:: https://user-images.githubusercontent.com/31141763/102669098-f6199d00-418d-11eb-9ae6-26c12d9c1231.gif\n\nSee `Configuring mlf-core <https://mlf_core.readthedocs.io/en/latest/config.html>`_\n\nlist\n----\nList all available mlf-core templates.\n\n.. figure:: https://user-images.githubusercontent.com/31141763/102668939-8d322500-418d-11eb-8b2c-acd895fc50e3.gif\n\nSee `Listing all templates <https://mlf_core.readthedocs.io/en/latest/list_info.html#list>`_.\n\ninfo\n----\nGet detailed information on a mlf-core template.\n\n.. figure:: https://user-images.githubusercontent.com/31141763/102669191-324cfd80-418e-11eb-9542-d2995b7318a9.gif\n\nSee `Get detailed template information <https://mlf_core.readthedocs.io/en/latest/list_info.html#info>`_.\n\ncreate\n------\nKickstart your deterministic machine laerning project with one of mlf-core's templates in no time.\n\n.. figure:: https://user-images.githubusercontent.com/31141763/102669143-1184a800-418e-11eb-853b-0deb0387efc6.gif\n\nSee `Create a project <https://mlf_core.readthedocs.io/en/latest/create.html>`_.\n\nlint\n----\nUse advanced linting to ensure your project always adheres to mlf-core's standards and stays deterministic.\n\n.. image:: https://user-images.githubusercontent.com/31141763/102668893-696edf00-418d-11eb-888e-822244a6f5dc.gif\n\nSee `Linting your project <https://mlf_core.readthedocs.io/en/latest/lint.html>`_\n\nbump-version\n------------\nBump your project version across several files.\n\n.. figure:: https://user-images.githubusercontent.com/31141763/102668987-aaff8a00-418d-11eb-9292-dc512f77f09b.gif\n\nSee `Bumping the version of an existing project  <https://mlf_core.readthedocs.io/en/latest/bump_version.html>`_.\n\nsync\n------\nSync your project with the latest mlf-core release to get the latest template features.\n\n.. figure:: https://user-images.githubusercontent.com/31141763/102669065-de421900-418d-11eb-9e1b-a76487d02b2a.gif\n\nSee `Syncing a project <https://mlf_core.readthedocs.io/en/latest/sync.html>`_.\n\nupgrade\n-------\nCheck whether you are using the latest mlf-core version and update automatically to benefit from the latest features.\n\nSee `<https://mlf_core.readthedocs.io/en/latest/upgrade.html>`_.\n\n\nCredits\n-------\n\nPrimary idea and main development by `Lukas Heumos <https://github.com/zethson/>`_. mlf-core is inspired by nf-core_.\nThis package was created with cookietemple_ based on a modified `audreyr/cookiecutter-pypackage`_ project template using cookiecutter_.\n\n.. _MLflow: https://mlflow.org\n.. _cookietemple: https://cookietemple.com\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT: http://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern_Python_Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _pip: https://pip.pypa.io/\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://mlf-core.readthedocs.io/en/latest/usage.html\n.. _cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`audreyr/cookiecutter-pypackage`: https://github.com/audreyr/cookiecutter-pypackage\n.. _nf-core: https://nf-co.re\n",
    'author': 'Lukas Heumos',
    'author_email': 'lukas.heumos@posteo.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mlf-core/mlf-core',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<3.10',
}


setup(**setup_kwargs)



import hashlib
import os
import random
import subprocess
import tempfile

import mlflow
import numpy as np
from rich import print

{%- if cookiecutter.language == "pytorch" %}
import torch

{%- elif cookiecutter.language == "tensorflow" %}
import tensorflow as tf

{%- endif %}


class MLFCore:
    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            print('Creating the object')
            cls._instance = super(MLFCore, cls).__new__(cls)
        return cls._instance

    @staticmethod
    def set_general_random_seeds(seed):
        os.environ['PYTHONHASHSEED'] = str(seed)  # Python general
        np.random.seed(seed)  # Numpy random
        random.seed(seed)  # Python random

    @staticmethod
    def log_system_intelligence(reports_output_dir: str):
        # Scoped import to prevent issues like RuntimeError: Numba cannot operate on non-primary CUDA context
        from system_intelligence.query import query_and_export

        print(f'[bold blue]Writing reports locally to {reports_output_dir}\n')
        print('[bold blue]Running system-intelligence')
        query_and_export(query_scope={'all'},
                         verbose=False,
                         export_format='json',
                         generate_html_table=True,
                         output=f'{reports_output_dir}/system_intelligence.json')
        print('[bold blue]Uploading system-intelligence report as a run artifact...')
        mlflow.log_artifacts(reports_output_dir, artifact_path='reports')

    @staticmethod
    def log_conda_environment(reports_output_dir: str):
        print('[bold blue]Exporting conda environment...')
        conda_env_filehandler = open(f'{reports_output_dir}/{{ cookiecutter.project_slug_no_hyphen }}_conda_environment.yml', 'w')
        subprocess.call(['conda', 'env', 'export', '--name', '{{ cookiecutter.project_slug_no_hyphen }}'], stdout=conda_env_filehandler)
        print('[bold blue]Uploading conda environment report as a run artifact...')
        mlflow.log_artifact(f'{reports_output_dir}/{{ cookiecutter.project_slug_no_hyphen }}_conda_environment.yml', artifact_path='reports')

{%- if cookiecutter.language == "pytorch" %}

    @staticmethod
    def set_pytorch_random_seeds(seed, num_gpus):
        torch.manual_seed(seed)
        torch.set_deterministic(True)
        if num_gpus > 0:
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)  # For multiGPU

{%- elif cookiecutter.language == "tensorflow" %}

    @staticmethod
    def set_tensorflow_random_seeds(seed):
        tf.random.set_seed(seed)
        tf.config.threading.set_intra_op_parallelism_threads = 1  # CPU only
        tf.config.threading.set_inter_op_parallelism_threads = 1  # CPU only
        os.environ['TF_DETERMINISTIC_OPS'] = '1'

{%- elif cookiecutter.language == "xgboost" %}

    @staticmethod
    def set_xgboost_random_seeds(seed, param):
        param['seed'] = seed

{%- elif cookiecutter.language == "xgboost_dask" %}

    @staticmethod
    def set_xgboost_dask_random_seeds(seed, param):
        param['seed'] = seed

{%- endif %}

    @classmethod
    def log_sys_intel_conda_env(cls):
        reports_output_dir = tempfile.mkdtemp()
        cls.log_system_intelligence(reports_output_dir)
        cls.log_conda_environment(reports_output_dir)

    @staticmethod
    def md5(fname: str):
        """Generate md5 sum for input file"""
        # Adding nosec (bandit) here, these isn't for security, just for tracking file integrity.
        hash_md5 = hashlib.md5()  # nosec
        with open(fname, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b''):
                hash_md5.update(chunk)

        md5sum = hash_md5.hexdigest()
        return md5sum

    @classmethod
    def get_md5_sums(cls, dir, max_files=None):
        """ Walk through directory and collect md5 sums """

        input_files = []
        for root, dirs, files in os.walk(dir, topdown=True):
            dirs.sort()
            files.sort()

            # Get maximum number of files to hash
            no_of_files = len(files)
            if max_files:
                no_of_files = max_files
            file_count = 0

            for elem in files:
                file_count += 1
                if file_count > no_of_files:
                    break

                elem = os.path.join(root, elem)
                elem_md5 = cls.md5(elem)
                input_files.append(elem_md5)

        input_files.sort()
        # Create temp file, write all the md5sums in it and hash the file
        # Gets deleted afterwards
        _, path = tempfile.mkstemp()
        try:
            with open(path, 'w') as tmp:
                tmp.writelines(input_files)
            dir_hash = cls.md5(path)
        finally:
            os.remove(path)

        return dir_hash

    @classmethod
    def log_input_data(cls, input_data: str, max_files=None):
        """
        Log input data by calculating a hash sum
        :param input_data: path to input data that should be logged
        :param max_files: maximum number of files to hash
        """
        print('[bold blue]Hashing input data...')
        if os.path.isdir(input_data):
            input_hash = cls.get_md5_sums(input_data, max_files=max_files)
        else:
            input_hash = cls.md5(input_data)
        mlflow.log_param("training_data_hash", input_data + "-" + input_hash)



from argparse import ArgumentParser

import pytorch_lightning as pl
import torch
from torch.autograd import Variable
from torch.nn import functional as F


class LightningMNISTClassifier(pl.LightningModule):
    def __init__(self, len_test_set: int, hparams: dict, **kwargs):
        """
        Initializes the network
        """
        super(LightningMNISTClassifier, self).__init__()
        self.hparams = hparams

        # mnist images are (1, 28, 28) (channels, width, height)
        self.optimizer = None
        self.conv1 = torch.nn.Conv2d(1, 32, 3, 1)
        self.conv2 = torch.nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = torch.nn.Dropout2d(0.25)
        self.fc1 = torch.nn.Linear(9216, 128)
        self.dropout2 = torch.nn.Dropout2d(0.25)
        self.fc2 = torch.nn.Linear(128, 10)
        self.args = kwargs
        self.len_test_set = len_test_set
        self.train_acc = pl.metrics.Accuracy()
        self.test_acc = pl.metrics.Accuracy()

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--num_workers', type=int, default=3, metavar='N', help='number of workers (default: 3)')
        parser.add_argument('--lr', type=float, default=0.01, help='learning rate (default: 0.01)')
        parser.add_argument('--training-batch-size', type=int, default=64, help='Input batch size for training')
        parser.add_argument('--test-batch-size', type=int, default=1000, help='Input batch size for testing')

        return parser

    def forward(self, x):
        """
        :param x: Input data

        :return: output - mnist digit label for the input image
        """
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2)
        x = torch.flatten(self.dropout1(x), 1)
        x = F.relu(self.fc1(x))
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)

        return output

    def cross_entropy_loss(self, logits, labels):
        """
        Initializes the loss function

        :return: output - Initialized cross entropy loss function
        """
        return F.nll_loss(logits, labels)

    def training_step(self, train_batch, batch_idx):
        """
        Training the data as batches and returns training loss on each batch

        :param train_batch: Batch data
        :param batch_idx: Batch indices

        :return: output - Training loss
        """
        x, y = train_batch
        logits = self.forward(x)
        loss = self.cross_entropy_loss(logits, y)
        self.train_acc(logits, y)
        self.log('train_acc', self.train_acc, on_step=True, on_epoch=True)
        return {'loss': loss}

    def training_epoch_end(self, training_step_outputs):
        """
        On each training epoch end, log the average training loss
        """
        train_avg_loss = torch.stack([train_output['loss'] for train_output in training_step_outputs]).mean()
        self.log('train_avg_loss', train_avg_loss, sync_dist=True)

    def test_step(self, test_batch, batch_idx):
        """
        Predicts on the test dataset to compute the current accuracy of the model.

        :param test_batch: Batch data
        :param batch_idx: Batch indices

        :return: output - Testing accuracy
        """

        x, y = test_batch
        output = self.forward(x)
        _, y_hat = torch.max(output, dim=1)
        self.test_acc(y_hat, y)
        self.log('test_acc', self.test_acc, on_step=False, on_epoch=True)
        # sum up batch loss
        data, target = Variable(x), Variable(y)  # noqa: F841
        test_loss = F.nll_loss(output, target, reduction='sum').data.item()
        # get the index of the max log-probability
        pred = output.data.max(1)[1]
        correct = pred.eq(target.data).sum()
        return {'test_loss': test_loss, 'correct': correct}

    def test_epoch_end(self, outputs):
        """
        Computes average test accuracy score

        :param outputs: outputs after every epoch end

        :return: output - average test loss
        """
        avg_test_loss = sum([test_output['test_loss'] for test_output in outputs]) / self.len_test_set
        test_correct = float(sum([test_output['correct'] for test_output in outputs]))
        self.log('avg_test_loss', avg_test_loss, sync_dist=True)
        self.log('test_correct', test_correct, sync_dist=True)

    def prepare_data(self):
        """
        Prepares the data for training and prediction
        """
        return {}

    def configure_optimizers(self):
        """
        Initializes the optimizer and learning rate scheduler

        :return: output - Initialized optimizer and scheduler
        """
        self.optimizer = torch.optim.Adam(self.parameters(), lr=self.args['lr'])
        self.scheduler = {
            'scheduler': torch.optim.lr_scheduler.ReduceLROnPlateau(
                self.optimizer, mode='min', factor=0.2, patience=2, min_lr=1e-6, verbose=True,
            ),
            'monitor': 'train_avg_loss',
        }
        return [self.optimizer], [self.scheduler]



import tensorflow as tf


def create_model(input_shape):
    model = tf.keras.Sequential([
        tf.keras.layers.Conv2D(32, 3, activation='relu', input_shape=input_shape),
        tf.keras.layers.Conv2D(64, 3, activation='relu'),
        tf.keras.layers.MaxPooling2D(2),
        tf.keras.layers.Dropout(0.25),
        tf.keras.layers.Flatten(),
        tf.keras.layers.Dropout(0.25),
        tf.keras.layers.Dense(64, activation='relu'),
        tf.keras.layers.Dense(10)
    ])

    return model



import os
from argparse import ArgumentParser

import mlflow
import pytorch_lightning as pl
from data_loading.data_loader import MNISTDataModule
from model.model import LightningMNISTClassifier
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import TensorBoardLogger
from rich import print

from mlf_core.mlf_core import MLFCore

if __name__ == "__main__":
    parser = ArgumentParser(description='PyTorch Autolog Mnist Example')
    parser.add_argument(
        '--general-seed',
        type=int,
        default=0,
        help='General random seed',
    )
    parser.add_argument(
        '--pytorch-seed',
        type=int,
        default=0,
        help='Random seed of all Pytorch functions',
    )
    parser.add_argument(
        '--log-interval',
        type=int,
        default=100,
        help='log interval of stdout',
    )
    parser = pl.Trainer.add_argparse_args(parent_parser=parser)
    parser = LightningMNISTClassifier.add_model_specific_args(parent_parser=parser)

    mlflow.autolog(1)
    # log conda env and system information
    MLFCore.log_sys_intel_conda_env()
    # parse cli arguments
    args = parser.parse_args()
    dict_args = vars(args)
    # store seed and number of gpus to make linter bit less restrict in terms of naming
    general_seed = dict_args['general_seed']
    pytorch_seed = dict_args['pytorch_seed']
    num_of_gpus = dict_args['gpus']
    MLFCore.set_general_random_seeds(general_seed)
    MLFCore.set_pytorch_random_seeds(pytorch_seed, num_of_gpus)

    if 'accelerator' in dict_args:
        if dict_args['accelerator'] == 'None':
            dict_args['accelerator'] = None
        elif dict_args['accelerator'] != 'ddp':
            print(f'[bold red]{dict_args["accelerator"]}[bold blue] currently not supported. Switching to [bold green]ddp!')
            dict_args['accelerator'] = 'ddp'

    dm = MNISTDataModule(**dict_args)

    # TODO MLF-CORE: Enable input data logging
    # MLFCore.log_input_data('data/')

    dm.prepare_data()
    dm.setup(stage='fit')
    model = LightningMNISTClassifier(len_test_set=len(dm.df_test), hparams=parser.parse_args(), **dict_args)
    model.log_every_n_steps = dict_args['log_interval']

    # check, whether the run is inside a Docker container or not
    if 'MLF_CORE_DOCKER_RUN' in os.environ:
        checkpoint_callback = ModelCheckpoint(filepath='/mlflow/tmp/mlruns', save_top_k=1, verbose=True, monitor='train_avg_loss', mode='min', prefix='',)
        trainer = pl.Trainer.from_argparse_args(args, checkpoint_callback=checkpoint_callback, default_root_dir='/data', logger=TensorBoardLogger('/data'))
        tensorboard_output_path = f'data/default/version_{trainer.logger.version}'
    else:
        checkpoint_callback = ModelCheckpoint(filepath=os.getcwd(), save_top_k=1, verbose=True, monitor='train_avg_loss', mode='min', prefix='',)
        trainer = pl.Trainer.from_argparse_args(args, checkpoint_callback=checkpoint_callback)
        tensorboard_output_path = f'{os.getcwd()}/lightning_logs/version_{trainer.logger.version}'

    trainer.deterministic = True
    trainer.benchmark = False
    trainer.log_every_n_steps = dict_args['log_interval']
    trainer.fit(model, dm)
    trainer.test()
    print(f'\n[bold blue]For tensorboard log, call [bold green]tensorboard --logdir={tensorboard_output_path}')
